# display.py
#
# Derek Devnich     30 April 2007
'''This separates the display components from the main method of Persephone.'''

import pygame
from pygame.locals import *
from persparms import gui_x, gui_y, label_y

# Define fonts
pygame.font.init()
symfont = pygame.font.SysFont('trebuchetms', label_y, True)

# Define colors
BLACK = (0,0,0)
WHITE = (255,255,255)
RED = (255,0,0)
GREEN = (0,255,0)
YELLOW = (255,255,50)
BLUE = (50,150,255)
GRAY = (150,150,150)
ORANGE = (255,100,0)

#-------------------------------------------------
# Define GUI elements
#-------------------------------------------------

class Label(object):
    '''Defines a free-standing text field.'''
    
    def __init__(self, screen, name, x, y, update_flag=True, label_type='normal'):
        self.name = name
        self.label_type = label_type
        self.update_flag = update_flag          # Update label on next iteration

        # Define location of the label.  A rect is defined as:
        # (x coordinate, y coordinate, width, height)
        self.rect = (x, y, gui_x*2, label_y)

        # Render text        
        self.text = symfont.render(name, True, WHITE, BLACK)

    def update_label(self, update_list, screen, label_type=None, new_label=None):
        '''Updates the color of the text based on its label_type,
        blits the text to the label's coordinates, and appends those
        coordinates to the update list.  This function gets called any
        time a label is added or changed.'''

        screen.fill(BLACK, self.rect)
        if new_label is not None:
            self.name = new_label
        if label_type is not None:
            self.label_type = label_type
        if self.label_type == 'normal':
            self.text = symfont.render(self.name, True, WHITE, BLACK)
        elif self.label_type == 'driver':
            self.text = symfont.render(self.name, True, YELLOW, BLACK)
        screen.blit(self.text, self.rect)
        update_list.append(self.rect)

    def update_location(self, screen, update_list, diff_y):
        '''Changes location of a unit.  The whole screen needs to be
        updated when this happens, by calling NetworkDisplay.update_area().'''

        screen.fill(BLACK, self.rect)
        self.rect = (self.rect[0], self.rect[1] + diff_y, self.rect[2], self.rect[3])
        update_list.append(self.rect)


class NodeDisplay(object):
    '''Represents a single unit in the network.'''
    
    def __init__(self, screen, unit, color, x, y, update_flag=True, offset=4):
        self.unit = unit
        self.color = color
        self.old_retrieved = False
        self.act = 0.0
        self.input = 0.0
        self.update_flag = True

        # Define locations of node and subfields
        self.rect = (x, y, gui_x, gui_y)
        self.text_rect = (x + offset, y + offset, gui_x - 10, 15)
        self.input_rect = (x + offset, self.text_rect[1] + self.text_rect[3] + offset, gui_x - 10, 3)
        self.act_rect = (x + offset, self.input_rect[1] + self.input_rect[3], gui_x - 10, 10)

        # Render content        
        self.border = pygame.draw.rect(screen, color, self.rect, 1)
        self.text = symfont.render(unit.name, True, color, BLACK)

    def draw_border(self, screen, update_list, retrieve_flag=False):
        '''Redraw border based on retrieved status of node.'''
        if retrieve_flag == False:
            screen.fill(BLACK, self.rect)
            screen.blit(self.text, self.text_rect)
            self.border = pygame.draw.rect(screen, self.color, self.rect, 1)
            update_list.append(self.rect)
        else:
            if self.old_retrieved != self.unit.retrieved:
                if self.unit.retrieved == True:
                    screen.fill(BLACK, self.rect)
                    screen.blit(self.text, self.text_rect)
                    self.border = pygame.draw.rect(screen, WHITE, self.rect, 5)
                elif self.unit.retrieved == False:
                    screen.fill(BLACK, self.rect)
                    screen.blit(self.text, self.text_rect)
                    self.border = pygame.draw.rect(screen, self.color, self.rect, 1)
                update_list.append(self.rect)
                self.old_retrieved = self.unit.retrieved

    def update_location(self, screen, update_list, diff_y):
        '''Changes location of a unit.  The whole screen needs to be
        updated when this happens, by calling NetworkDisplay.update_area().'''

        screen.fill(BLACK, self.rect)
        update_list.append(self.rect)
              
        self.rect = (self.rect[0], self.rect[1] + diff_y, self.rect[2], self.rect[3])
        self.text_rect = (self.text_rect[0], self.text_rect[1] + diff_y, self.text_rect[2], self.text_rect[3])
        self.input_rect = (self.input_rect[0], self.input_rect[1] + diff_y, self.input_rect[2], self.input_rect[3])
        self.act_rect = (self.act_rect[0], self.act_rect[1] + diff_y, self.act_rect[2], self.act_rect[3])

        update_list.append(self.rect)
        
    def update_act(self, screen, update_list):
        '''Updates the activation bar of a unit, based on its new or
        current activation.  This function needs to be called from a
        loop that iterates through all the units in the network.'''

        self.act = self.unit.act
        
        screen.fill(BLACK, self.act_rect)
        current_act = (self.act_rect[0], self.act_rect[1], self.act_rect[2] * self.act, self.act_rect[3])
        screen.fill(self.color, current_act)
        update_list.append(self.act_rect)

    def update_input(self, screen, update_list):
        '''Updates the input bar of a unit, based on its new or
        current input.  This function needs to be called from a loop
        that iterates through all the units in the network.

        This function will have to be scaled to deal with the various input
        sources.  Generally, source should be a number in the range [0:1].'''

        ### self.input = self.unit....

        screen.fill(BLACK, self.input_rect)
        current_input = (self.input_rect[0], self.input_rect[1], self.input_rect[2] * self.input,
                         self.input_rect[3])
        screen.fill(GRAY, current_input)
        update_list.append(self.input_rect)

    def update_mode(self, network_display):
        if hasattr(self.unit, 'mode'):
            if self.unit.mode == 'child':
                self.color = network_display.colors['obj']
            else:
                self.color = network_display.colors['prop']


#-------------------------------------------------
# Define GUI control objects
#-------------------------------------------------

class NetworkDisplay(object):
    '''Defines a container that controls the location, drawing, and
    display of all elements of the graphical portion of the GUI.'''
    
    def __init__(self, y=0):
        # These lists contain all Label and NodeDisplay objects
        self.nodes = {'target':{'label':[],'group':[],'prop':[],'rf':[],'pred':[],'obj':[]},
                      'ltm':{'label':[],'group':[],'prop':[],'rf':[],'pred':[],'obj':[]},
                      'source':{'label':[],'group':[],'prop':[],'rf':[],'pred':[],'obj':[]},
                      'schema':{'label':[],'group':[],'prop':[],'rf':[],'pred':[],'obj':[]},
                      'world':{'label':[],'group':[],'prop':[],'rf':[],'pred':[],'obj':[]},
                      'semantic':{'label':[],'pred':[],'obj':[],'episode':[], 'link':[]},
                      'status':{'input':[],'speed':[],'pause':[]}
                      }

        # Keywords used to access self.nodes in the correct order
        self.kwds = [['target','label'],['target','group'],['target','prop'],['target','rf'],
                     ['target','pred'],['target','obj'],
                     ['ltm','label'],['ltm','group'],['ltm','prop'],['ltm','rf'],
                     ['ltm','pred'],['ltm','obj'],
                     ['source','label'],['source','group'],['source','prop'],['source','rf'],
                     ['source','pred'],['source','obj'],
                     ['schema','label'],['schema','group'],['schema','prop'],['schema','rf'],
                     ['schema','pred'],['schema','obj'],
                     ['world','label'],['world','group'],['world','prop'],['world','rf'],
                     ['world','pred'],['world','obj'],
                     ['semantic','label'],['semantic','pred'],['semantic','obj'],['semantic','episode'],
                     ['semantic','link'],
                     ['status','input'],['status','speed'],['status','pause']
                     ]

        self.colors = {'group':ORANGE,'prop':YELLOW,'rf':GREEN,'pred':BLUE,'obj':RED,'semantic':WHITE}
        self.driver_border_rect = None
        self.driver_border = None

        # Status field locations
        self.status = {'input':0,'speed':2,'pause':4}

        # Possible sources of input
        self.input = ['net input','top-down','bottom-up','lateral','outprop','hebb','hebb weight','None']

    def purge_display(self, screen):
        for kwd in self.kwds:
            self.nodes[kwd[0]][kwd[1]] =[]
        screen.fill(BLACK)
        pygame.display.flip()

    def set_driver_border(self, screen, update_list, driver):
        # Erase old driver border by calling draw_network() before this,
        # with draw_border(retrieve_flag=False)
        if self.driver_border_rect is not None:
            pygame.draw.rect(screen, BLACK, self.driver_border_rect, 3)
            update_list.append(self.driver_border_rect)
        
        # Calculate new driver border
        if driver in self.nodes:
            x = self.nodes[driver]['label'][0].rect[0]
            y = self.nodes[driver]['label'][0].rect[1]
            width = pygame.display.get_surface().get_width()
            height = self.nodes[driver]['obj'][-1].rect[1] + gui_y - y

            self.driver_border_rect = (x, y, width, height)
            self.driver_border = pygame.draw.rect(screen, YELLOW, self.driver_border_rect, 3)
            update_list.append(self.driver_border_rect)
        
    def update_area(self, screen, update_list):
        '''This function recalculates the position of all display
        elements.  It iterates through each list of units in
        self.nodes, positioning the first element of each list
        directly below the last row of the previous list.  If the
        previous list is empty, the function looks up through
        self.nodes until it finds a list that isn't.'''

        for kwd in self.kwds:
            if self.kwds.index(kwd) > 0:
                prior = self.kwds[self.kwds.index(kwd)-1]
                all_empty = False
                while len(self.nodes[prior[0]][prior[1]]) == 0:
                    if self.kwds.index(prior) > 0:
                        prior = self.kwds[self.kwds.index(prior)-1]
                    else:
                        all_empty = True
                        break
                if all_empty == False:
                    try:
                        if self.nodes[kwd[0]][kwd[1]][0].rect[1] <= self.nodes[prior[0]][prior[1]][-1].rect[1]:
                            new_y = self.nodes[prior[0]][prior[1]][-1].rect[1] + gui_y
                            old_y = self.nodes[kwd[0]][kwd[1]][0].rect[1]
                            diff_y = new_y - old_y
                            if diff_y != 0:
                                for node in self.nodes[kwd[0]][kwd[1]]:
                                    node.update_location(screen, update_list, diff_y)
                            #screen.fill(BLACK)
                    except IndexError:
                        continue
        pygame.display.update(update_list)

    def add_status(self, screen, status_text, kwd=['status','input']):
        '''Adds a status text field to the bottom of the screen.'''
        
        height = pygame.display.get_surface().get_height()
        y = height - label_y - 4
        x = self.status[kwd[1]] * gui_x
        self.nodes[kwd[0]][kwd[1]].append(Label(screen, kwd[1] + ': ' + status_text, x, y,
                                                update_flag=True, label_type='normal'))

    def add_label(self, screen, name, update_list, kwd=['target','label']):
        '''Adds a label to the screen.  This is used mainly for section titles.'''
        
        home = kwd
        all_empty = False
        while len(self.nodes[kwd[0]][kwd[1]]) == 0:
            if self.kwds.index(kwd) > 0:
                kwd = self.kwds[self.kwds.index(kwd)-1]
            else:
                all_empty = True
                break
        if all_empty == False:
            x = 0
            y = self.nodes[kwd[0]][kwd[1]][-1].rect[1] + gui_y
            self.nodes[home[0]][home[1]].append(Label(screen, name, x, y, update_flag=True, label_type='normal'))
            self.update_area(screen, update_list)
        else:
            self.nodes[home[0]][home[1]].append(Label(screen, name, 0, 0, update_flag=True, label_type='normal'))

    def add_node(self, unit, screen, update_list, kwd=['target','group'], color=GRAY):
        '''Adds a DisplayNode to the screen.'''
        
        width = pygame.display.get_surface().get_width()

        # If the current list already has units, cue off of the last one
        try:
            (x,y,q,r) = self.nodes[kwd[0]][kwd[1]][-1].rect
            # Add the unit to the same row, if it will fit
            if width >= x + 2*gui_x:
                self.nodes[kwd[0]][kwd[1]].append(NodeDisplay(screen, unit, color, x+gui_x, y,
                                                              update_flag=True))
            else:
                # Start new row, which requires repainting everything
                x = 0
                y = self.nodes[kwd[0]][kwd[1]][-1].rect[1] + gui_y
                self.nodes[kwd[0]][kwd[1]].append(NodeDisplay(screen, unit, color, x, y, update_flag=True))
                self.update_area(screen, update_list)

        # If the list is empty, go up the lists until you find one that isn't
        except IndexError:
            home = kwd
            all_empty = False
            while len(self.nodes[kwd[0]][kwd[1]]) == 0:
                if self.kwds.index(kwd) > 0:
                    kwd = self.kwds[self.kwds.index(kwd)-1]
                else:
                    all_empty = True
                    break
            if all_empty == False:
                x = 0
                y = self.nodes[kwd[0]][kwd[1]][-1].rect[1] + gui_y
                self.nodes[home[0]][home[1]].append(NodeDisplay(screen, unit, color, x, y, update_flag=True))
                self.update_area(screen, update_list)
            else:
                self.nodes[home[0]][home[1]].append(NodeDisplay(screen, unit, color, 0, 0, update_flag=True))

    def clear_analog(self, analog_name):
        '''Clears display elements out of the designated analog'''

        for list_type, node_list in self.nodes[analog_name].iteritems():
            if list_type != 'label':
                del node_list[:]

    def add_analog(self, analog, screen, update_list):
        '''Creates a complete set of display nodes for the analog.
        Call this when an analog has been cleared and then copied
        to.'''

        for group in analog.groups:
            self.add_node(group, screen, update_list, [analog.name, 'group'], self.colors['group'])
        for prop in analog.props:
            self.add_node(prop, screen, update_list, [analog.name, 'prop'], self.colors['prop'])
        for rf in analog.rfbindings:
            self.add_node(rf, screen, update_list, [analog.name, 'rf'], self.colors['rf'])
        for pred in analog.preds:
            self.add_node(pred, screen, update_list, [analog.name, 'pred'], self.colors['pred'])
        for obj in analog.objs:
            self.add_node(obj, screen, update_list, [analog.name, 'obj'], self.colors['obj'])

    def add_semantics(self, pred_semantics, obj_semantics, episode_semantics, link_semantics,
                      screen, update_list):
        self.add_label(screen, 'Semantics', update_list, ['semantic','label'])
        for pred_sem in pred_semantics:
            self.add_node(pred_sem, screen, update_list, ['semantic', 'pred'], self.colors['semantic'])
        for obj_sem in obj_semantics:
            self.add_node(obj_sem, screen, update_list, ['semantic', 'obj'], self.colors['semantic'])
        for group_sem in episode_semantics:
            self.add_node(group_sem, screen, update_list, ['semantic', 'episode'], self.colors['semantic'])
        for group_sem in link_semantics:
            self.add_node(group_sem, screen, update_list, ['semantic', 'link'], self.colors['semantic'])
                
    def draw_network(self, screen, update_list, flag_needed=False):
        '''Draws the complete state of the network to the screen.'''
        
        screen.fill(BLACK)
        for analog_name, analog in self.nodes.iteritems():
            for nodes in analog.itervalues():
                for node in nodes:
                    if flag_needed == False or (flag_needed == True and node.update_flag == True):
                        if type(node) == Label:
                            screen.blit(node.text, node.rect)
                            update_list.append(node.rect)
                        elif type(node) == NodeDisplay:
                            node.update_mode(self)
                            node.draw_border(screen, update_list, retrieve_flag=False)
                            screen.blit(node.text, node.text_rect)
                            node.update_act(screen, update_list)
                        node.update_flag = False
        pygame.display.update(update_list)
        update_list = []
