'''network.py           14 February 2006

A collection of methods that builds and connects Persephone nodes.'''

import perstypes

def create_analog(analog_name, analogs):
    if analog_name not in analogs:  
        analogs[analog_name] = perstypes.Analog(analog_name)
    return analogs[analog_name]
        
def create_semantic(semantic_name, unit_semantics, semantics):
    '''Creates a semantic unit and inserts it into the global "semantics" dict.  It also adds
    the unit to the "unit_semantics" dict that the symparser uses to attach semantics units
    to OP units.'''

    semantic = [i for i in semantics if i.name == semantic_name]
    if len(semantic) > 0:
        unit_semantics.append(semantic[0])
    else:
        semantic = perstypes.Semantic(semantic_name)
        semantics.append(semantic)
        unit_semantics.append(semantic)
        
def create_op(optype, analog, opname, sequence, unit_semantics=None, inferred=False):
    '''Creates an Obj or Pred unit, attaches semantics, and inserts it into an analog.  The call
    from the symparser looks like this:

    create_op('pred', analogs[analog_name], role_name, all_pred_semantics)
    create_op('obj', analogs[analog_name], obj_name, unit_semantics)

    If "unit_semantics" isn't specified, the method infers new unit with no semantics.'''

    if optype == 'obj':
        if unit_semantics is not None and inferred == False:
            op = perstypes.Obj(opname)
            for semantic in unit_semantics:
                op.semantic[semantic] = 1.0
                semantic.op[op] = 1.0
        elif inferred == True:
            opname = opname + '*'
            op = perstypes.Obj(opname)
            op.inferred = inferred
            
        analog.objs.append(op)
        op.sequence = sequence
        sequence.objs.append(op)

    if optype == 'pred':
        if unit_semantics is not None and inferred == False:
            op = perstypes.Pred(opname)
            for semantic in unit_semantics:
                op.semantic[semantic] = 1.0
                semantic.op[op] = 1.0
        elif inferred == True:
            opname = opname + '*'
            op = perstypes.Pred(opname)
            op.inferred = inferred

        analog.preds.append(op)
        op.sequence = sequence
        sequence.preds.append(op)

    return op

def create_rf(analog, rf_name, sequence, prop=None, pred=None, obj=None, inferred=False):
    '''Creates an RFBinding unit, attaches it to other units, and inserts in into an analog.'''

    if inferred == False:
        rf = perstypes.RFBinding(rf_name)
                   
        # connect to Pred unit
        if pred is not None:
            pred.rf.append(rf)
            rf.pred = pred
        # connect to parent Prop unit
        if prop is not None:
            prop.rf.append(rf)
            rf.prop = prop
        # connect to Object...
        if type(obj) is perstypes.Obj:
            obj.rf.append(rf)
            rf.obj = obj
        # ...or child Prop
        elif type(obj) is perstypes.Prop:
            obj.parent_rf.append(rf)
            rf.obj = obj
            
    elif inferred == True:
        rf_name = rf_name + '*'
        rf = perstypes.RFBinding(rf_name)
        rf.inferred = inferred
        rf.infant = True

    analog.rfbindings.append(rf)
    rf.sequence = sequence
    sequence.rfbindings.append(rf)
        
    return rf

def create_prop(analog, prop_name, sequence, inferred=False):
    '''Creates a prop unit and inserts in into an analog.'''
    
    if inferred == False:
        prop = perstypes.Prop(prop_name)
    elif inferred == True:
        prop_name = prop_name + '*'
        prop = perstypes.Prop(prop_name)
        prop.inferred = inferred

    analog.props.append(prop)
    sequence.props.append(prop)
    prop.sequence = sequence
    
    return prop

def create_group(analog, group_name, sequence, group_type=None, group_semantics=None, state_props=None,
                 action_props=None, episodes=None, value=0.0, reward=0.0, causal_power=0.0,
                 inferred=False, direction=None):
    
    if inferred == False:        
        if group_type == 'episode':
            group = perstypes.Episode(group_name)
            if state_props is not None:
                group.state = state_props
                for prop in state_props:
                    prop.episode_role = 'state'
                    prop.group.append(group)
            if action_props is not None:
                group.state.extend(action_props)
                for prop in action_props:
                    prop.episode_role = 'action'
                    prop.group.append(group)
        elif group_type == 'link':
            group = perstypes.Link(group_name)

        if group_semantics is not None:
            for semantic in group_semantics:
                group.semantic[semantic] = 1.0
                semantic.op[group] = 1.0

        sequence.ordered_groups.append(group)
        group.sequence = sequence

    elif inferred == True:
        group_name = group_name + '*'
        if group_type == 'episode':
            group = perstypes.Episode(group_name)
        elif group_type == 'link':
            group = perstypes.Link(group_name)
        group.inferred = inferred
        group.infant = True

        # Insert groups into sequence in the correct order; depends on direction...
        # ordered_groups = [C.1, L.1, E.1, C.2, L.2, E.2....]
        # inserting at the correct end of the list is handled here;
        # ordering within set will be handled by connect_new_units()
        if direction == 'explain':
            sequence.ordered_groups.insert(0,group)
        else:
            sequence.ordered_groups.append(group)
        group.sequence = sequence

    group.value = value
    group.reward = reward
    group.causal_power = causal_power
    analog.groups.append(group)

    return group

def create_hebbs(analogs, driver, hebbs, theta_r):
    '''Creates hebbs from each recipient to the driver.  If a prop in the recipient gets
    "retrieved" (i.e., it has a hebb created to it), all of its subsidiary units are also
    retrieved.  In addition, OP units can be independently retrieved.

    Hebbs are only created to driver units that are in the current phase set.'''

    ### => Check mappable and phase set; mappable is only set for target, and will always be True for source
    
    for analog_name, analog in analogs.iteritems():
        if analog.analog_type == 'recipient' and analog.top_down_ok == True:

            # Retrieve props in the recipient
            for prop in analog.props:
                if (prop.act > theta_r or prop.retrieved == True) and prop.mappable == True:
                    prop.get_hebbset(driver)
                    prop.retrieved = True
                    for driver_prop in analogs[driver].props:
                        if driver_prop.phase_set_flag == True and driver_prop.mappable == True:
                            if prop.mode == 'child' and driver_prop.mode == 'child':
                                driver_prop.get_hebbset(analog_name)
                                prop.get_hebb(analog_name, driver, driver_prop, hebbs)

                            # Parent props in the recicpient cue recall of all of their constituents
                            elif prop.mode == 'parent' and driver_prop.mode == 'parent':
                                driver_prop.get_hebbset(analog_name)
                                prop.get_hebb(analog_name, driver, driver_prop, hebbs)
                                prop.retrieved = True

                                # Retrieve owned RFs of parent props in the recipient
                                # Assume mappable recipient props have mappable RFs, don't check
                                for rf in prop.rf:
                                    rf.get_hebbset(driver)
                                    rf.retrieved = True
                                    for driver_rf in analogs[driver].rfbindings:
                                        if driver_rf.phase_set_flag == True and driver_rf.mappable == True:
                                            driver_rf.get_hebbset(analog_name)
                                            rf.get_hebb(analog_name, driver, driver_rf, hebbs)

                                    # Retrieve owned Preds of owned RFs
                                    if rf.pred is not None:
                                        rf.pred.get_hebbset(driver)
                                        rf.pred.retrieved = True
                                        for driver_pred in analogs[driver].preds:
                                            if driver_pred.phase_set_flag == True and driver_pred.mappable == True:
                                                driver_pred.get_hebbset(analog_name)
                                                rf.pred.get_hebb(analog_name, driver, driver_pred, hebbs)

                                    # Retrieve owned Objs or Child Props of owned RFs
                                    if rf.obj is not None:
                                        rf.obj.get_hebbset(driver)
                                        rf.obj.retrieved = True
                                        if type(rf.obj) is perstypes.Obj:
                                            for driver_obj in analogs[driver].objs:
                                                if driver_obj.phase_set_flag == True and driver_obj.mappable == True:
                                                    driver_obj.get_hebbset(analog_name)
                                                    rf.obj.get_hebb(analog_name, driver, driver_obj, hebbs)
                                        elif type(rf.obj) is perstypes.Prop:
                                            for driver_prop in analogs[driver].props:
                                                if driver_prop.phase_set_flag == True and driver_prop.mode == 'child':  # and driver_prop.mappable == True:
                                                    driver_prop.get_hebbset(analog_name)
                                                    rf.obj.get_hebb(analog_name, driver, driver_prop, hebbs)

            # Retrieve Preds in the recipient
            for pred in analog.preds:
                if (pred.act > theta_r or pred.retrieved == True) and pred.mappable == True:
                    pred.get_hebbset(driver)
                    pred.retrieved = True
                    for driver_pred in analogs[driver].preds:
                        if driver_pred.phase_set_flag == True and driver_pred.mappable == True:
                            driver_pred.get_hebbset(analog_name)
                            pred.get_hebb(analog_name, driver, driver_pred, hebbs)

            # Retrieve Objs in the recipient
            for obj in analog.objs:
                if (obj.act > theta_r or obj.retrieved == True) and obj.mappable == True:
                    obj.get_hebbset(driver)
                    obj.retrieved = True
                    for driver_obj in analogs[driver].objs:
                        if driver_obj.phase_set_flag == True and driver_obj.mappable == True:
                            driver_obj.get_hebbset(analog_name)
                            obj.get_hebb(analog_name, driver, driver_obj, hebbs)

            # Retrieve groups in the recipient; probably will never be invoked, but hey...
            for group in analog.groups:
                if (group.act > theta_r or group.retrieved == True) and group.mappable == True:
                    group.get_hebbset(driver)
                    group.retrieved = True
                    for driver_group in analogs[driver].groups:
                        if driver_group.phase_set_flag == True and driver_group.mappable == True:

                            ### No groups to map to in driver
                            if type(driver_group) == perstypes.Link and type(group) == perstypes.Link:
                                driver_group.get_hebbset(analog_name)
                                group.get_hebb(analog_name, driver, driver_group, hebbs)
                            elif type(driver_group) == perstypes.Episode and type(group) == perstypes.Episode:
                                driver_group.get_hebbset(analog_name)
                                group.get_hebb(analog_name, driver, driver_group, hebbs)
 
def find_consistent_hebbs(analogs, driver, hebbs, eval_flag=False):
    '''Creates excitatory connections between mutually consistent hebbs that connect units
    at adjoining levels of the hierarchy (e.g., RFs and their OPs).

    Doesn't detect split mappings, or incomplete sets of mappings.'''

    ### => Don't check mappable; DO check phase set in driver

    consistent_mapping = True

    # Iterate through props in the driver's phase set
    for prop in analogs[driver].props:
        if prop.phase_set_flag == True or eval_flag == True:
            for target_analog, hebbset in prop.hebbsets.iteritems():
                target_props = [hebbs[hebb].owners[target_analog] for hebb in hebbset]

                # Iterate through prop's child RFs
                for rf in prop.rf:
                    # Phase set membership ought to be guaranteed....
                    if rf.phase_set_flag == True or eval_flag == True:
                        # Create list of hebbs out of this unit, and list of units to which they go.
                        # The lists are in the same order.
                        rf_hebbset = rf.hebbsets[target_analog]
                        target_rfs = [hebbs[hebb].owners[target_analog] for hebb in rf_hebbset]

                        # Find consistent hebbs
                        for target_rf in target_rfs:
                            if target_rf.prop is not None and target_rf.prop in target_props:
                                if eval_flag == False:
                                    rf_hebb = rf_hebbset[target_rfs.index(target_rf)]
                                    prop_hebb = hebbset[target_props.index(target_rf.prop)]
                                    hebbs[rf_hebb].connections[prop_hebb] = 1.0
                                    hebbs[prop_hebb].connections[rf_hebb] = 1.0
                            else:
                                if eval_flag == True:
                                    consistent_mapping = False
                                    print 'Consistency failure for:', prop.name, rf.name, 'to', target_analog, target_rf.prop.name, target_rf.name

    # Iterate through child props in the driver's phase set
    for prop in analogs[driver].props:
        if (prop.phase_set_flag == True or eval_flag == True) and prop.mode == 'child':
            for target_analog, hebbset in prop.hebbsets.iteritems():
                target_props = [hebbs[hebb].owners[target_analog] for hebb in hebbset]

                # Iterate through prop's parent RFs
                for rf in prop.parent_rf:
                    if rf.phase_set_flag == True or eval_flag == True:
                        rf_hebbset = rf.hebbsets[target_analog]
                        target_rfs = [hebbs[hebb].owners[target_analog] for hebb in rf_hebbset]

                        # Find consistent hebbs
                        for target_rf in target_rfs:
                            if target_rf.obj is not None and target_rf.obj in target_props:
                                if eval_flag == False:
                                    rf_hebb = rf_hebbset[target_rfs.index(target_rf)]
                                    prop_hebb = hebbset[target_props.index(target_rf.obj)]
                                    hebbs[rf_hebb].connections[prop_hebb] = 1.0
                                    hebbs[prop_hebb].connections[rf_hebb] = 1.0
                            else:
                                if eval_flag == True:
                                    consistent_mapping = False
                                    print 'Consistency failure for:', rf.name, prop.name, 'to', target_analog, target_rf.name, target_rf.obj.name

    # Iterate through preds in the driver's phase set
    for pred in analogs[driver].preds:
        if pred.phase_set_flag == True or eval_flag == True:
            for target_analog, hebbset in pred.hebbsets.iteritems():
                target_preds = [hebbs[hebb].owners[target_analog] for hebb in hebbset]
            
                # Iterate through pred's parent RFs
                for rf in pred.rf:
                    if rf.phase_set_flag == True or eval_flag == True:
                        if target_analog in rf.hebbsets:
                            rf_hebbset = rf.hebbsets[target_analog]
                            target_rfs = [hebbs[hebb].owners[target_analog] for hebb in rf_hebbset]

                            # Find consistent hebbs
                            for target_rf in target_rfs:
                                if target_rf.pred is not None and target_rf.pred in target_preds:
                                    if eval_flag == False:
                                        #print 'Found overlapping target:', target_rf.pred.name
                                        rf_hebb = rf_hebbset[target_rfs.index(target_rf)]
                                        pred_hebb = hebbset[target_preds.index(target_rf.pred)]
                                        hebbs[rf_hebb].connections[pred_hebb] = 1.0
                                        hebbs[pred_hebb].connections[rf_hebb] = 1.0
                                else:
                                    if eval_flag == True:
                                        consistent_mapping = False
                                        print 'Consistency failure for:', rf.name, pred.name,'to', target_analog, target_rf.name, target_rf.pred.name

    # Iterate through objs in the driver's phase set
    for obj in analogs[driver].objs:
        if obj.phase_set_flag == True or eval_flag == True:
            for target_analog, hebbset in obj.hebbsets.iteritems():
                target_objs = [hebbs[hebb].owners[target_analog] for hebb in hebbset]

                # Iterate through obj's parent RFs
                for rf in obj.rf:
                    if rf.phase_set_flag == True or eval_flag == True:
                        if target_analog in rf.hebbsets:
                            rf_hebbset = rf.hebbsets[target_analog]
                            target_rfs = [hebbs[hebb].owners[target_analog] for hebb in rf_hebbset]

                            # Find consistent hebbs
                            for target_rf in target_rfs:
                                if target_rf.obj is not None and target_rf.obj in target_objs:
                                    if eval_flag == False:
                                        rf_hebb = rf_hebbset[target_rfs.index(target_rf)]
                                        obj_hebb = hebbset[target_objs.index(target_rf.obj)]
                                        hebbs[rf_hebb].connections[obj_hebb] = 1.0
                                        hebbs[obj_hebb].connections[rf_hebb] = 1.0
                                else:
                                    if eval_flag == True:
                                        consistent_mapping = False
                                        print 'Consistency failure for:', rf.name, obj.name,'to', target_analog, target_rf.name, target_rf.obj.name

    # Iterate through props in the driver's phase set
    for prop in analogs[driver].props:
        if prop.phase_set_flag == True or eval_flag == True:
            for target_analog, hebbset in prop.hebbsets.iteritems():
                target_props = [hebbs[hebb].owners[target_analog] for hebb in hebbset]

                # Iterate through prop's parent groups
                for group in prop.group:
                    # Skip non-mappable groups in target
                    if group.phase_set_flag == True and target_analog in group.hebbsets:
                        group_hebbset = group.hebbsets[target_analog]
                        target_groups = [hebbs[hebb].owners[target_analog] for hebb in group_hebbset]

                        # Find consistent hebbs
                        for target_group in target_groups:
                            for prop_2 in target_group.state:
                                if prop_2 in target_props:
                                    if eval_flag == False:
                                        group_hebb = group_hebbset[target_groups.index(target_group)]
                                        prop_hebb = hebbset[target_props.index(prop_2)]
                                        hebbs[group_hebb].connections[prop_hebb] = 1.0
                                        hebbs[prop_hebb].connections[group_hebb] = 1.0
                                else:                                   
                                    pass
    ##                                if eval_flag == True:
    ##                                    consistent_mapping = False
    ##                                    print 'Consistency failure for:', prop.name, group.name,'to', target_analog, prop_2.name, target_group.name

    # Iterate through episode groups in the driver's phase set
    for episode in analogs[driver].groups:
        if type(episode) == perstypes.Episode and episode.phase_set_flag == True:  # or type(episode) == perstypes.Effect:
            for target_analog, hebbset in episode.hebbsets.iteritems():
                target_episodes = [hebbs[hebb].owners[target_analog] for hebb in hebbset]

                # Iterate through episode's causal link groups
                #for link in episode.link:
                if episode.link.phase_set_flag == True:
                    link_hebbset = episode.link.hebbsets[target_analog]
                    target_links = [hebbs[hebb].owners[target_analog] for hebb in link_hebbset]
                                    
                    # Find consistent hebbs
                    for target_link in target_links:
                        for episode_2 in target_link.episode:
                            if episode_2 in target_episodes:
                                if eval_flag == False:
                                    link_hebb = link_hebbset[target_links.index(target_link)]
                                    episode_hebb = hebbset[target_episodes.index(episode_2)]
                                    hebbs[link_hebb].connections[episode_hebb] = 1.0
                                    hebbs[episode_hebb].connections[link_hebb] = 1.0

    return consistent_mapping

def kill_hebbs(analogs, hebbs, theta_k):
    '''Called at the end of the phase set to kill off hebbs with a weight > theta_k.  Removes
    the hebb from the global "hebbs" dict, and hebb's name from the hebbset of any unit
    containing it.'''
    
    for hebbname, hebb in hebbs.items():
        if hebb.perm_wt < theta_k:
            for analog in analogs.itervalues():
                if analog.analog_type == 'driver' or analog.analog_type == 'recipient':
                    for pred in analog.preds:
                        for hebbset in pred.hebbsets.itervalues():
                            if hebbname in hebbset:
                                hebbset.remove(hebbname)
                    for obj in analog.objs:
                        for hebbset in obj.hebbsets.itervalues():
                            if hebbname in hebbset:
                                hebbset.remove(hebbname)
                    for rf in analog.rfbindings:
                        for hebbset in rf.hebbsets.itervalues():
                            if hebbname in hebbset:
                                hebbset.remove(hebbname)
                    for prop in analog.props:
                        for hebbset in prop.hebbsets.itervalues():
                            if hebbname in hebbset:
                                hebbset.remove(hebbname)
                    for group in analog.groups:
                        for hebbset in group.hebbsets.itervalues():
                            if hebbname in hebbset:
                                hebbset.remove(hebbname)
            for h in hebbs.itervalues():
                if hebbname in h.connections:
                    del h.connections[hebbname]
            del hebbs[hebbname]

def get_importance(analog):

    '''Calculates importance of the various Token units.  This value is used by license_learning() to
    determine whether inference is warranted.'''

    if analog.analog_type == 'recipient' and analog.top_down_ok == True:
        for pred in analog.preds:
            pred.temp_importance = 0.0
            
            try:
                total_prop_imp = sum(rf.prop.importance for rf in pred.rf)
            except AttributeError:
                print pred.name, [(rf.name, rf.prop) for rf in pred.rf]
            pred.temp_importance = pred.importance + total_prop_imp
        for obj in analog.objs:
            obj.temp_importance = 0.0
            total_prop_imp = sum(rf.prop.importance for rf in obj.rf)
            obj.temp_importance = pred.importance + total_prop_imp


def license_learning(analog, pi_m):
    '''Determines whether the importance-weighted mapping quality from recipient to driver is
    sufficient to license inference.'''

    if analog.analog_type == 'recipient' and analog.top_down_ok == True:

        prop_weighted = sum(prop.importance * (prop.max_hebb_w - prop.second_max_hebb_w)
                            for prop in analog.props if prop.mappable == True)
        prop_total = sum(prop.importance for prop in analog.props if prop.mappable == True)

        pred_weighted = sum(pred.temp_importance * (pred.max_hebb_w - pred.second_max_hebb_w)
                            for pred in analog.preds if pred.mappable == True)
        pred_total = sum(pred.temp_importance for pred in analog.preds if pred.mappable == True)

        obj_weighted = sum(obj.temp_importance * (obj.max_hebb_w - obj.second_max_hebb_w)
                           for obj in analog.objs if obj.mappable == True)
        obj_total = sum(obj.temp_importance for obj in analog.objs if obj.mappable == True)

        weighted = sum([prop_weighted, pred_weighted, obj_weighted])
        total = sum([prop_total, pred_total, obj_total])

        if weighted/total > pi_m:
            analog.ready_to_learn = True

def infer_new_units(analogs, direction, driver, hebbs,
                    network_display, screen, update_list, pi_i,j):
    '''Infers a new unit of a given type (in the recipient) when:
    1) No units of that type in the recipient are receiving hebb excitation, and
    2) Some proportion of units of that type in the recipient are receiving hebb inhibition.

    When a new unit is inferred, it automatically gets a hebb connection of 1.0 to the most active
    unit of its type in the driver, specified by "driver_*_name".'''

    for analog_name, analog in analogs.iteritems():
        if analog.analog_type == 'recipient' and analog.ready_to_learn == True and analog.top_down_ok == True:
            sequence = analog.sequences[0]
            
            # Check for new Preds
            pred_units = [unit for unit in analog.preds if type(unit) == perstypes.Pred
                          and unit.mappable == True]
            zero_units = [unit for unit in analog.preds if unit.hebb_excite <= 0 and unit.mappable == True]
            inhib_units = [unit for unit in analog.preds if unit.hebb_inhib > 0 and unit.mappable == True]
            #print len(zero_units), len(analog.preds), len(inhib_units), len(zero_units)/pi_i
            if len(zero_units) == len(pred_units) and len(inhib_units) >= len(zero_units)/pi_i:
                
                if analog.driver_pred is not None:
                    unit = create_op('pred', analog, analog.driver_pred.name, sequence, inferred=True)
                    unit.get_hebbset(driver)
                    hebb = unit.get_hebb(analog_name, driver, analog.driver_pred, hebbs)
                    hebbs[hebb].perm_wt = 1.0
                    unit.net_input = 1.0
                    unit.act = 1.0

                    if screen is not None:
                        network_display.add_node(unit, screen, update_list, [analog.name, 'pred'],
                                                 network_display.colors['pred'])
                        network_display.draw_network(screen, update_list)

            # Check for new Objs
            obj_units = [unit for unit in analog.objs if type(unit) == perstypes.Obj and unit.mappable == True]
            zero_units = [unit for unit in analog.objs if unit.hebb_excite <= 0 and unit.mappable == True]
            inhib_units = [unit for unit in analog.objs if unit.hebb_inhib > 0 and unit.mappable == True]
            if len(zero_units) == len(obj_units) and len(inhib_units) >= len(zero_units)/pi_i:
                if analog.driver_obj is not None:
                    unit = create_op('obj', analog, analog.driver_obj.name, sequence, inferred=True)
                    unit.get_hebbset(driver)
                    hebb = unit.get_hebb(analog_name, driver, analog.driver_obj, hebbs)
                    hebbs[hebb].perm_wt = 1.0
                    unit.net_input = 1.0
                    unit.act = 1.0
                    
                    if screen is not None:
                        network_display.add_node(unit, screen, update_list, [analog.name, 'obj'],
                                                 network_display.colors['obj'])
                        network_display.draw_network(screen, update_list)

            # Check for new RFs
            rf_units = [unit for unit in analog.rfbindings if type(unit) == perstypes.RFBinding
                        and unit.mappable == True]
            zero_units = [unit for unit in analog.rfbindings if unit.hebb_excite <= 0 and unit.mappable == True]
            inhib_units = [unit for unit in analog.rfbindings if unit.hebb_inhib > 0 and unit.mappable == True]
            if len(zero_units) == len(rf_units) and len(inhib_units) >= len(zero_units)/pi_i:
                if analog.driver_rf is not None:
                    unit = create_rf(analog, analog.driver_rf.name, sequence, inferred=True)
                    unit.get_hebbset(driver)
                    hebb = unit.get_hebb(analog_name, driver, analog.driver_rf, hebbs)
                    hebbs[hebb].perm_wt = 1.0
                    unit.net_input = 1.0
                    unit.act = 1.0

                    if screen is not None:
                        network_display.add_node(unit, screen, update_list, [analog.name, 'rf'],
                                                 network_display.colors['rf'])
                        network_display.draw_network(screen, update_list)
                   
            # Check for new parent Props
            parent_units = [unit for unit in analog.props if unit.mode != 'child' and unit.mappable == True]
            zero_units = [unit for unit in analog.props if unit.hebb_excite <= 0 and unit.mode != 'child'
                          and unit.mappable == True]
            inhib_units = [unit for unit in analog.props if unit.hebb_inhib > 0 and unit.mode != 'child'
                           and unit.mappable == True]
            if len(zero_units) == len(parent_units) and len(inhib_units) >= len(zero_units)/pi_i:
                if analog.driver_prop is not None:
                    unit = create_prop(analog, analog.driver_prop.name, sequence, inferred=True)
                    unit.mode = analog.driver_prop.mode
                    unit.episode_role = analog.driver_prop.episode_role
                    unit.get_hebbset(driver)
                    hebb = unit.get_hebb(analog_name, driver, analog.driver_prop, hebbs)
                    hebbs[hebb].perm_wt = 1.0
                    unit.net_input = 1.0
                    unit.act = 1.0
                    
                    if screen is not None:
                        network_display.add_node(unit, screen, update_list, [analog.name, 'prop'],
                                                 network_display.colors['prop'])
                        network_display.draw_network(screen, update_list)
                    
            # Check for new child Props
            child_units = [unit for unit in analog.props if unit.mode != 'parent' and unit.mappable == True]
            zero_units = [unit for unit in analog.props if unit.hebb_excite <= 0 and unit.mode != 'parent'
                          and unit.mappable == True]
            inhib_units = [unit for unit in analog.props if unit.hebb_inhib > 0 and unit.mode != 'parent'
                           and unit.mappable == True]
            if len(zero_units) == len(child_units) and len(inhib_units) >= len(zero_units)/pi_i:
                if analog.driver_c_prop is not None:
                    unit = create_prop(analog, analog.driver_c_prop.name, sequence, inferred=True)
                    unit.mode = analog.driver_c_prop.mode
                    unit.episode_role = analog.driver_c_prop.episode_role
                    unit.get_hebbset(driver)
                    hebb = unit.get_hebb(analog_name, driver, analog.driver_c_prop, hebbs)
                    hebbs[hebb].perm_wt = 1.0
                    unit.net_input = 1.0
                    unit.act = 1.0

                    if screen is not None:
                        network_display.add_node(unit, screen, update_list, [analog.name, 'prop'],
                                                 network_display.colors['prop'])
                        network_display.draw_network(screen, update_list)

            # Check for new Episode groups
            episode_units = [unit for unit in analog.groups if type(unit) == perstypes.Episode
                             and unit.mappable == True]
            zero_units = [unit for unit in analog.groups if unit.hebb_excite <= 0 and
                          type(unit) == perstypes.Episode and unit.mappable == True]
            inhib_units = [unit for unit in analog.groups if unit.hebb_inhib > 0 and
                           type(unit) == perstypes.Episode and unit.mappable == True]
            if len(zero_units) == len(episode_units) and len(inhib_units) >= len(zero_units)/pi_i:
                if analog.driver_episode is not None:
                    unit = create_group(analog, analog.driver_episode.name, sequence, group_type = 'episode',
                                        value = analog.driver_episode.value, inferred=True,
                                        direction = direction)
                    unit.get_hebbset(driver)
                    hebb = unit.get_hebb(analog_name, driver, analog.driver_episode, hebbs)
                    hebbs[hebb].perm_wt = 1.0
                    unit.net_input = 1.0
                    unit.act = 1.0

                    if screen is not None:
                        network_display.add_node(unit, screen, update_list, [analog.name, 'group'],
                                                 network_display.colors['group'])
                        network_display.draw_network(screen, update_list)

            # Check for new Causal Link groups
            link_units = [unit for unit in analog.groups if type(unit) == perstypes.Link
                          and unit.mappable == True]
            zero_units = [unit for unit in analog.groups if unit.hebb_excite <= 0 and
                          type(unit) == perstypes.Link and unit.mappable == True]
            inhib_units = [unit for unit in analog.groups if unit.hebb_inhib > 0 and
                           type(unit) == perstypes.Link and unit.mappable == True]
            if len(zero_units) == len(link_units) and len(inhib_units) >= len(zero_units)/pi_i:
                if analog.driver_link is not None:
                    unit = create_group(analog, analog.driver_link.name, sequence, group_type = 'link',
                                        causal_power = analog.driver_link.causal_power, inferred=True,
                                        direction = direction)
                    unit.get_hebbset(driver)
                    hebb = unit.get_hebb(analog_name, driver, analog.driver_link, hebbs)
                    hebbs[hebb].perm_wt = 1.0
                    #unit.hebb_excite = 1.0
                    unit.net_input = 1.0
                    unit.act = 1.0

                    if screen is not None:
                        network_display.add_node(unit, screen, update_list, [analog.name, 'group'],
                                                 network_display.colors['group'])
                        network_display.draw_network(screen, update_list)


def connect_new_units(analogs, direction, driver, pred_semantics, obj_semantics,
                      episode_semantics, link_semantics, theta_c):
    '''Connects newly-inferred units to other units in the analog (inferred or not).  OP units
    are responsible for finding their semantics; RFBinding units are responsible for finding
    OP units, props, and child props (if necessary).

    The method always selects the most active Token unit of a given type as its candidate for
    connection (assuming that other critera are also met).'''
    
    for analog in analogs.itervalues():
        if analog.analog_type == 'recipient' and analog.ready_to_learn == True and analog.top_down_ok == True:
            sequence = analog.sequences[0]

            #--------------------------------------
            # Connect inferred pred to semantics
            #--------------------------------------
            
            # Get most active inferred pred
            candidate_pred = None
            for pred in analog.preds:
                if pred.inferred == True and pred.act > theta_c:
                    if candidate_pred is not None:
                        if pred.act > candidate_pred.act:
                            candidate_pred = pred
                    else:
                        candidate_pred = pred

            # Connect pred to semantics
            if candidate_pred is not None:
                for semantic in pred_semantics:
                    if semantic.act > theta_c:
                        if candidate_pred not in semantic.op:
                            semantic.op[candidate_pred] = 0.0
                        if semantic not in candidate_pred.semantic:
                            candidate_pred.semantic[semantic] = 0.0

            #--------------------------------------
            # Connect inferred obj to semantics
            #--------------------------------------
            
            # Get most active inferred obj
            candidate_obj = None
            for obj in analog.objs:
                if obj.inferred == True and obj.act > theta_c:
                    if candidate_obj is not None:
                        if obj.act > candidate_obj.act:
                            candidate_obj = obj
                    else:
                        candidate_obj = obj

            # Connect obj to semantics
            if candidate_obj is not None:
                for semantic in obj_semantics:
                    if semantic.act > theta_c:
                        if candidate_obj not in semantic.op:
                            semantic.op[candidate_obj] = 0.0
                        if semantic not in candidate_obj.semantic:
                            candidate_obj.semantic[semantic] = 0.0

                #print 'Most active obj:', candidate_obj.name

            #--------------------------------------
            # Connect inferred groups to semantics
            #--------------------------------------

            # Get most active inferred Episode Group
            candidate_episode = None
            for group in analog.groups:
                if type(group) == perstypes.Episode:    # or type(group) == perstypes.Effect:
                    if group.inferred == True and group.act > theta_c:
                        if candidate_episode is not None:
                            if group.act > candidate_episode.act:
                                candidate_episode = group
                        else:
                            candidate_episode = group
                            
            # Connect episode to semantics
            if candidate_episode is not None:
                for semantic in episode_semantics:
                    if semantic.act > theta_c:
                        if candidate_episode not in semantic.op:
                            semantic.op[candidate_episode] = 0.0
                        if semantic not in candidate_episode.semantic:
                            candidate_episode.semantic[semantic] = 0.0

            # Get most active inferred Link Group
            candidate_link = None
            for group in analog.groups:
                if type(group) == perstypes.Link and group.inferred == True and group.act > theta_c:
                    if candidate_link is not None:
                        if group.act > candidate_link.act:
                            candidate_link = group
                    else:
                        candidate_link = group

            # Connect link to semantics
            if candidate_link is not None:
                for semantic in link_semantics:
                    if semantic.act > theta_c:
                        if candidate_link not in semantic.op:
                            semantic.op[candidate_link] = 0.0
                        if semantic not in candidate_link.semantic:
                            candidate_link.semantic[semantic] = 0.0

            #--------------------------------------
            # Connect inferred RF
            #--------------------------------------

            # Get most active inferred RF
            candidate_rf = None
            for rf in analog.rfbindings:
                if rf.inferred == True and rf.infant == True and rf.act > theta_c:
                    if candidate_rf is not None:
                        if rf.act > candidate_rf.act:
                            candidate_rf = rf
                    else:
                        candidate_rf = rf

            # Look for parent prop
            candidate_prop = None
            for prop in analog.props:
                if prop.act > theta_c and prop.mode == 'parent':
                    if candidate_prop is not None:
                        if prop.act > candidate_prop.act:
                            candidate_prop = prop
                    else:
                        candidate_prop = prop
            if candidate_prop is not None and candidate_rf is not None and candidate_rf.prop is None:
                candidate_prop.rf.append(candidate_rf)
                candidate_rf.prop = candidate_prop

            # Get most active pred of any kind
            candidate_pred = None
            for pred in analog.preds:
                if pred.act > theta_c:
                    if candidate_pred is not None:
                        if pred.act > candidate_pred.act:
                            candidate_pred = pred
                    else:
                        candidate_pred = pred
            if candidate_rf is not None and candidate_pred is not None and candidate_rf.pred is None:
                candidate_pred.rf.append(candidate_rf)
                candidate_rf.pred = candidate_pred

            # Get most active obj of any kind
            candidate_obj = None
            for obj in analog.objs:
                if obj.act > theta_c:
                    if candidate_obj is not None:
                        if obj.act > candidate_obj.act:
                            candidate_obj = obj
                    else:
                        candidate_obj = obj

            # Look for obj or child prop
            #if candidate_rf is not None and candidate_rf.obj is None:
            # Check if child props are more active than most active obj
            for prop in analog.props:
                if prop.mode == 'child' and prop.act > theta_c:
                    if candidate_obj is not None:
                        if prop.act > candidate_obj.act:
                            candidate_obj = prop
                    else:
                        candidate_obj = prop

            # Connect obj/child prop with RF        
            if candidate_rf is not None and candidate_obj is not None and candidate_rf.obj is None:
                if type(candidate_obj) is perstypes.Obj:
                    candidate_obj.rf.append(candidate_rf)
                    candidate_rf.obj = candidate_obj
                elif type(candidate_obj) is perstypes.Prop:
                    candidate_obj.parent_rf.append(candidate_rf)
                    candidate_rf.obj = candidate_obj
                #print 'Obj is:', candidate_obj.name
           
            # Check if RF is grown up
            for rf in analog.rfbindings:
                if rf.infant == True:
                    if rf.prop is not None and rf.pred is not None and rf.obj is not None:
                        rf.infant = False
                        #print 'RF grows up:', rf.name

            #--------------------------------------
            # Connect inferred Group
            #--------------------------------------

            '''This code may cause problems if more than one Link
            group is available at a time.  The fact that RFs only take
            one unit of each type restricts the connection
            possibilities of OPs and props.  Link groups may be
            similarly restricted - exactly two episodes, one with a
            Cause semantic and one with an Effect semantic - but until
            we introduce this restriction, connections may go awry.'''         

            # Get most active Episode Group
            candidate_episode = None
            for group in analog.groups:
                if type(group) == perstypes.Episode and group.act > theta_c:
                    if candidate_episode is not None:
                        if group.act > candidate_episode.act:
                            candidate_episode = group
                    else:
                        candidate_episode = group
                       
            if candidate_episode is not None and candidate_prop is not None:
                if candidate_prop not in candidate_episode.state:
                    candidate_prop.group.append(candidate_episode)
                    candidate_episode.state.append(candidate_prop)
                    #print 'Prop attached:', candidate_episode.name, candidate_prop.name, [(i.name, i.act) for i in analogs['target'].groups]

            # Get most active Link group
            candidate_link = None
            for group in analog.groups:
                # Only take link groups that hold less than 2 episodes
                if type(group) == perstypes.Link and group.act > theta_c:
                    #if len(group.episode) < 2:   #group.cause is None or group.effect is None:
                    if candidate_link is not None:
                        if group.act > candidate_link.act:
                            candidate_link = group
                    else:
                        candidate_link = group

            # Connect link to episode                
            if candidate_episode is not None and candidate_link is not None:
                if candidate_episode.link is None and len(candidate_link.episode) < 2:
                    candidate_episode.link = candidate_link
                    candidate_link.episode.append(candidate_episode)
                    candidate_episode.infant = False
                if len(candidate_link.episode) == 2:
                    candidate_link.infant = False

            # When link is fully connected, sort its set in place in the sequence
            if candidate_link is not None and len(candidate_link.episode) == 2:
##            if candidate_link is not None and candidate_link.cause is not None and candidate_link.effect is not None:
                # Sorting sets
                cause = None
                effect = None
                for episode in candidate_link.episode:
                    cause_s = [i for i in episode.semantic.iterkeys() if i.name == 'cause']
                    effect_s = [i for i in episode.semantic.iterkeys() if i.name == 'effect']
                    if any(cause_s) and not any(effect_s):
                        cause = episode
                    elif not any(cause_s) and any(effect_s):
                        effect = episode
                if cause is not None and effect is not None:
                    E = sequence.ordered_groups.index(effect)
                    C = sequence.ordered_groups.index(cause)
                    if E < C:
                        sequence.ordered_groups[C], sequence.ordered_groups[E] = sequence.ordered_groups[E], sequence.ordered_groups[C]

                    L = sequence.ordered_groups.index(candidate_link)
                    C = sequence.ordered_groups.index(cause)
                    if L < C:
                        sequence.ordered_groups[C], sequence.ordered_groups[L] = sequence.ordered_groups[L], sequence.ordered_groups[C]

                    L = sequence.ordered_groups.index(candidate_link)
                    E = sequence.ordered_groups.index(effect)
                    if L > E:
                        sequence.ordered_groups[E], sequence.ordered_groups[L] = sequence.ordered_groups[L], sequence.ordered_groups[E]

                    # Check for other groups intermingling set
                    L = sequence.ordered_groups.index(candidate_link)
                    C = sequence.ordered_groups.index(cause)
                    E = sequence.ordered_groups.index(effect)
                    if (L - C) > 1 or (E - L) > 1:
                        print 'Inferred groups out of order in sequence.ordered_groups'
                        print '  source:', __file__, '.connect_new_units()'
    

def update_pred_semantic_weights(analogs, mu):
    '''Updates connection weights between semantics and inferred pred units.'''

    for analog in analogs.itervalues():
        if analog.analog_type == 'recipient' and analog.ready_to_learn == True and analog.top_down_ok == True:
            for pred in analog.preds:
                if pred.inferred == True:
                    for semantic, weight in pred.semantic.iteritems():
                        delta_w = mu * (semantic.act - weight) * pred.act
                        if delta_w < 0:
                            delta_w = delta_w * 0.5
                        new_weight = weight + delta_w
                        pred.semantic[semantic] = new_weight
                        semantic.op[pred] = new_weight

def update_obj_semantic_weights(analogs, mu):
    '''Updates connection weights between semantics and inferred obj units.'''

    for analog in analogs.itervalues():
        if analog.analog_type == 'recipient' and analog.ready_to_learn == True and analog.top_down_ok == True:
            for obj in analog.objs:
                if obj.inferred == True:
                    for semantic, weight in obj.semantic.iteritems():
                        delta_w = (mu * (semantic.act - weight)) * obj.act
                        if delta_w < 0:
                            delta_w = delta_w * 0.5
                        new_weight = weight + delta_w
                        obj.semantic[semantic] = new_weight
                        semantic.op[obj] = new_weight

def update_group_semantic_weights(analogs, mu):
    '''Updates connection weights between semantics and inferred obj units.'''

    for analog in analogs.itervalues():
        if analog.analog_type == 'recipient' and analog.ready_to_learn == True and analog.top_down_ok == True:
            for group in analog.groups:
                if group.inferred == True:
                    for semantic, weight in group.semantic.iteritems():
                        delta_w = (mu * (semantic.act - weight)) * group.act
                        if delta_w < 0:
                            delta_w = delta_w * 0.5
                        new_weight = weight + delta_w
                        group.semantic[semantic] = new_weight
                        semantic.op[group] = new_weight


