'''perstypes.py

Data types for the model, along with their attributes and methods.  All units
know their own name.'''

import itertools, random, os
from persparms import *

random.seed(os.urandom(99))

class Sequence(object):
    '''This is container for a complete episode.  It contains lists of all the
    objects composing a complete episode sequence.  The group list is ordered,
    and can be used to set firing order.  This should greatly alleviate problems
    with constructing and retrieving episodes.'''
    def __init__(self):
        self.ordered_groups = []
        self.props = []
        self.rfbindings = []
        self.preds = []
        self.objs = []
        self.retrieved = False

#--------------------------------------
# Hebb classes
#--------------------------------------
'''A hebb is a mapping connection between units of like types in different analogs.  The hebbs
coming out of a particular unit compete among themselves to become the most heavily weighted
connection to a given analog (the hebbs pointing to one analog don't compete with hebbs
pointing to a second analog.'''

class HebbDict(dict):
    '''The HebbDict (instantiated as "hebbs") is an extended dict that contains hebb units.'''
    
    def __init__(self):
        dict.__init__(self)
        self.hebb_count = 0


class Hebb(object):
    '''A hebb connects exactly two units.  This information is stored in "self.owners" when the
    hebb is created, as {'analog1_name':unit1, 'analog2_name':unit2}.'''
    
    def __init__(self, name):
        self.name = name
        self.owners = {}
        self.connections = {}   # consistent hebbs in {hebb_name: weight} format

        self.w_delta = 0.0
        self.perm_wt = 0.0
        self.buffer_delta = 0.0
        self.buffer = 0.0
        self.div_max = 1.0
        self.sub_max = 0.0

        self.excitation = 0.0
        self.inhibition = 0.0
        self.total_input = 0.0
        self.act = 0.0
        #self.update_flag = False    # used to determine if consistent hebbs have been updated
            
    def get_buffer_delta(self):
        '''Gets the change in the buffer based on the activation of the connected Token units.'''
        
        units = self.owners.values()
        self.buffer_delta = units[0].act * units[1].act

    def update_buffer(self):
        self.buffer = self.buffer + self.buffer_delta

    def get_max_buffer(self, hebbs):
        '''Returns the maximum buffer value of the hebbs belonging to the two hebbsets owned
        by the current hebb's Token units.  This value is used for divisive normaliziation of
        hebbs.'''
        
        analog = self.owners.keys()
        units = self.owners.values()

        try:
            buffers = [hebbs[i].buffer for i in units[0].hebbsets[analog[1]]]
            buffers_1 = [hebbs[i].buffer for i in units[1].hebbsets[analog[0]]]
        except KeyError:
            print analog[0], units[0].name, units[0].hebbsets[analog[1]]

        buffers.extend(buffers_1)
        if len(buffers) > 0 and max(buffers) > 0:
            self.div_max = max(buffers)
        else:
            self.div_max = 1.0

    def apply_max_buffer(self):
        '''Applies div_max to the buffers of each hebb.'''

        self.buffer = self.buffer/self.div_max

    def get_second_max_buffer(self, hebbs):
        '''Returns the maximum buffer value of the hebbs belonging to the two hebbsets owned
        by the current hebb's Token units, not including the current hebb.  This value is used
        for subtractive normaliziation of hebbs.'''
        
        analog = self.owners.keys()
        units = self.owners.values()

        buffers = [hebbs[i].buffer for i in units[0].hebbsets[analog[1]] if i !=self.name]
        buffers_1 = [hebbs[i].buffer for i in units[1].hebbsets[analog[0]] if i !=self.name]

        buffers.extend(buffers_1)
        if len(buffers) > 0:
            self.sub_max = max(buffers)
        else:
            self.sub_max = 0.0

    def get_second_max_act(self, hebbs):
        '''Returns the maximum activation value of the hebbs belonging to the two hebbsets owned
        by the current hebb's Token units, not including the current hebb.  This value is used
        for inhibition of inconsistent hebbs.'''
        
        analog = self.owners.keys()
        units = self.owners.values()

        act = [hebbs[i].act for i in units[0].hebbsets[analog[1]] if i != self.name]
        act_1 = [hebbs[i].act for i in units[1].hebbsets[analog[0]] if i != self.name]

        act.extend(act_1)
        if len(act) > 0:
            self.inhibition = max(act)
        else:
            self.inhibition = 0.0

    def get_consistent_act(self, hebbs):
        '''Gets the activations of consistent hebbs.  This version of the method is used in
        conjunction with the Green/Hummel/Iberri non-neural consistent hebbs method.'''

        self.excitation = sum(hebbs[i].act for i in self.connections.iterkeys())


    def get_total_input(self, rho_he, rho_hi):
        '''Gets total excitatory and inhibitory input to this hebb.

        The original form of this function multiplied self.excitation by 1 and
        self.inhibition by 10.  This led to bad behavior in the case of
        ambiguous mappings.  The algorithm does not tolerate ambiguity, and
        instead prefers to drive mappings to 1.  In an ambiguous situation, this
        results in the RFs within a prop mapping to several different props's
        RFs.'''
        
        self.total_input = self.buffer + rho_he * self.excitation - rho_hi * self.inhibition

    def update_activation(self, gamma_h, delta_h, theta_h):

        settled_flag = True
        old_act = self.act
        
        act_delta = gamma_h * self.total_input * (1 - abs(self.act)) - delta_h * self.act
        new_act = self.act + act_delta

        # Activation limited to [0,1] instead of [-1,1].
        
        if 1.0 >= new_act >= 0.0:
            self.act = new_act
        elif new_act > 1.0:
            self.act = 1.0
        elif new_act < 0.0:
            self.act = 0.0

        if abs(old_act - self.act) > theta_h:
            settled_flag = False
        return settled_flag

## Old weight change method for hebbs, based on competition by subtractive normalization
##    def get_w_delta(self, eta):
##        '''Gets the weight change of the current hebb.'''
##        
##        self.w_delta = eta * (1.1 - self.perm_wt) * (self.buffer - self.sub_max)

    def get_w_delta(self, eta):
        '''Gets the weight change of the current hebb.'''
        
        self.w_delta = eta * (1.1 - self.perm_wt) * self.act

    def update_perm_wt(self):
        perm_wt = self.perm_wt + self.w_delta
        if 0.0 <= perm_wt <= 1.0:
            self.perm_wt = perm_wt
        elif perm_wt > 1.0:
            self.perm_wt = 1.0
        else:
            self.perm_wt = 0.0        

    def flush(self):
        '''Sets buffers and deltas == 0.0'''

        self.buffer_delta = 0.0
        self.w_delta = 0.0
        self.buffer = 0.0
        self.div_max = 1.0
        self.sub_max = 0.0
        
        self.excitation = 0.0
        self.inhibition = 0.0
        self.total_input = 0.0
        self.act = 0.0
        self.connections.clear()


#--------------------------------------
# Token class and its descendents
#--------------------------------------
'''A general note about ownership of symmetrical connections within an analog:
Each unit knows about its connections to the units above it and below it.  This
information is stored as object references; all weights are assumed to be 1.'''

class Token(object):
    '''This is the base class for all Token units.'''

    def __init__(self):
        self.sequence = None            # Sequence object that collects this token
        self.mappable = True
        self.inferred = False
        self.infant = False             # Only used by RFs and Groups
        self.retrieved = False
        self.phase_set_flag = False
        self.importance = 1.0
        
        # Excitor fields
        self.bu_input = 0.0
        self.td_input = 0.0
        self.within_class_input = 0.0
        self.outprop_input = 0.0
        self.hebb_excite = 0.0
        self.hebb_inhib = 0.0
        self.max_hebb_w = 0.0
        self.max_hebb_id = None     # name of the max hebb from this recipient unit to the driver
        self.second_max_hebb_w = 0.0
        self.total_hebb_input = 0.0
        self.net_input = 0.0
        self.act_delta = 0.0
        self.act = 0.0
        self.old_act = 0.0
        
        # Hebb fields
        '''Hebbsets contain the names of hebbs going from this unit to units in another analog:
        {'a2':['h1','h2','h3']}'''
        
        self.hebbsets = {}

    def set_mappable(self, value=True):
        self.mappable = value

    #-------------------------------------------------
    # Hebb creation and and hebb-based input methods
    #-------------------------------------------------

    def get_hebbset(self, t_analog_name):
        '''Returns a unit's hebb set for a target analog, creating it if necessary.'''
        
        if t_analog_name in self.hebbsets:
            return self.hebbsets[t_analog_name]
        else:  
            self.hebbsets[t_analog_name] = []
            return self.hebbsets[t_analog_name]

    def get_hebb(self, recipient, driver, driver_unit, hebbs):
        '''Returns the hebb connecting a unit to one in a target analog, creating it if
        neccessary.  This function should be called from the POV of the RECIPIENT.

        Each hebb should only connnect units of like type (pred, obj, etc).  "get_hebb"
        should be called by a function that checks unit types.  This function does NOT
        do any error-checking to see if the hebbset exists.

        "driver" and "recipient" are the names of the respective analogs.  "driver_unit" is
        the Token unit in the driver analog that this unit is mapping to.'''

        shared_hebb = set(self.hebbsets[driver]).intersection(set(driver_unit.hebbsets[recipient]))
        if len(shared_hebb) > 0:
            hebbname = shared_hebb.pop()
        else:
            hebbname = 'h' + str(hebbs.hebb_count + 1)
            hebbs[hebbname] = Hebb(hebbname)
            hebbs.hebb_count = hebbs.hebb_count + 1
            hebbs[hebbname].owners = {recipient:self, driver:driver_unit}
            self.hebbsets[driver].append(hebbname)
            driver_unit.hebbsets[recipient].append(hebbname)
        return hebbname

    def get_hebb_excitation(self, analog, driver, hebbs):
        '''Gets hebb-based excitatory input.  This method checks to see if this unit
        has hebb connection to the currently firing driver unit of the same type.
        The driver unit activation is multiplied by 3 times the shared weight in order to
        counterbalance the global inhibition from "get_hebb_inhib".'''
        
        hebbset = self.get_hebbset(driver)
        shared_w = 0.0

        if type(self) is Pred:
            for hebb in hebbset:
                if hebbs[hebb].owners[driver] is analog.driver_pred:
                    shared_w = hebbs[hebb].perm_wt
            self.hebb_excite = analog.driver_pred_act * (3 * shared_w)

        elif type(self) is Obj:
            for hebb in hebbset:
                if hebbs[hebb].owners[driver] is analog.driver_obj:
                    shared_w = hebbs[hebb].perm_wt
            self.hebb_excite = analog.driver_obj_act * (3 * shared_w)
            
        elif type(self) is RFBinding:
            for hebb in hebbset:
                if hebbs[hebb].owners[driver] is analog.driver_rf:
                    shared_w = hebbs[hebb].perm_wt
            self.hebb_excite = analog.driver_rf_act * (3 * shared_w)

        elif type(self) is Prop and self.mode == 'parent':
            for hebb in hebbset:
                if hebbs[hebb].owners[driver] is analog.driver_prop:
                    shared_w = hebbs[hebb].perm_wt
            self.hebb_excite = analog.driver_prop_act * (3 * shared_w)

        elif type(self) is Prop and self.mode == 'neutral':
            shared_p_w = 0.0
            shared_c_w = 0.0
            for hebb in hebbset:
                if hebbs[hebb].owners[driver] is analog.driver_prop:
                    shared_p_w = hebbs[hebb].perm_wt
                if hebbs[hebb].owners[driver] is analog.driver_c_prop:
                    shared_c_w = hebbs[hebb].perm_wt
            p_excite = analog.driver_prop_act * (3 * shared_p_w)
            c_excite = analog.driver_c_prop_act * (3 * shared_c_w)
            self.hebb_excite = p_excite + c_excite
            
        elif type(self) is Prop and self.mode == 'child':
            for hebb in hebbset:
                if hebbs[hebb].owners[driver] is analog.driver_c_prop:
                    shared_w = hebbs[hebb].perm_wt
            self.hebb_excite = analog.driver_c_prop_act * (3 * shared_w)
            
        elif type(self) is Link:
            for hebb in hebbset:
                if hebbs[hebb].owners[driver] is analog.driver_link:
                    shared_w = hebbs[hebb].perm_wt
            self.hebb_excite = analog.driver_link_act * (3 * shared_w)
        elif type(self) is Episode:
            for hebb in hebbset:
                if hebbs[hebb].owners[driver] is analog.driver_episode:
                    shared_w = hebbs[hebb].perm_wt
            self.hebb_excite = analog.driver_episode_act * (3 * shared_w)

    def get_max_hebb_w(self, driver, hebbs):
        '''Gets the maximum hebb weight and 2nd maximum hebb weight coming into this
        recipient unit.'''

        hebbset = self.get_hebbset(driver)
        max_hebb_w = 0.0
        max_hebb_id = None
        second_max = 0.0
        
        hebb_w = [hebbs[hebb].perm_wt for hebb in hebbset]
        if len(hebb_w) > 0:
            max_hebb_id = hebbset[hebb_w.index(max(hebb_w))]
            hebb_w.sort()
            max_hebb_w = hebb_w.pop()
            if len(hebb_w) > 0:
                second_max = hebb_w.pop()
        self.max_hebb_id = max_hebb_id
        self.max_hebb_w = max_hebb_w
        self.second_max_hebb_w = second_max


    def get_hebb_inhib(self, analog):
        '''Returns the inhibitory portion of the total hebb input for checking for inference.
        Units are globally inhibited by the max hebb connecting units of their type.'''

        if type(self) is Pred:
            self.hebb_inhib = analog.driver_pred_act * (self.max_hebb_w + analog.driver_pred_hebb)
        elif type(self) is Obj:
            self.hebb_inhib = analog.driver_obj_act * (self.max_hebb_w + analog.driver_obj_hebb) 
        elif type(self) is RFBinding:
            self.hebb_inhib = analog.driver_rf_act * (self.max_hebb_w + analog.driver_rf_hebb)
        elif type(self) is Prop and self.mode == 'parent':
            self.hebb_inhib = analog.driver_prop_act * (self.max_hebb_w + analog.driver_prop_hebb)
        elif type(self) is Prop and self.mode == 'neutral':
            p_hebb_inhib = analog.driver_prop_act * (self.max_hebb_w + analog.driver_prop_hebb)
            c_hebb_inhib = analog.driver_c_prop_act * (self.max_hebb_w + analog.driver_c_prop_hebb)
            self.hebb_inhib = p_hebb_inhib + c_hebb_inhib
        elif type(self) is Prop and self.mode == 'child':
            self.hebb_inhib = analog.driver_c_prop_act * (self.max_hebb_w + analog.driver_c_prop_hebb)
        elif type(self) is Link:
            self.hebb_inhib = analog.driver_link_act * (self.max_hebb_w + analog.driver_link_hebb)
        elif type(self) is Episode:
            self.hebb_inhib = analog.driver_episode_act * (self.max_hebb_w + analog.driver_episode_hebb)

    def get_total_hebb_input(self, hebb_bias, action_flag):
        '''Returns the total hebb input into this unit.'''

        self.total_hebb_input = hebb_bias * self.hebb_excite - hebb_bias * self.hebb_inhib
    

    #-------------------------------------
    # Net input and activation methods
    #-------------------------------------

    def get_driver_net_input(self):
        '''Returns the net input to this driver unit.'''

        self.net_input = self.bu_input + self.td_input

    def get_recipient_net_input(self):
        '''Returns the net input to this recipient unit.'''

        self.net_input = (self.bu_input + self.td_input) + self.total_hebb_input - self.within_class_input - self.outprop_input

    def get_act_delta(self, gamma, delta):
        '''Returns change in activation for the current timestep using a leaky integrator.'''

        self.act_delta = gamma * self.net_input * (1 - self.act) - delta * self.act

    def update_act(self):
        self.old_act = self.act
        if self.mappable == False:  # and self.phase_set_flag == False:
            self.act = 0.0
        else:
            if self.act + self.act_delta < 0:
                self.act = 0.0
            elif self.act + self.act_delta > 1:
                self.act = 1.0
            else:
                self.act = self.act + self.act_delta


class OP(Token):
    def __init__(self, name):
        Token.__init__(self)
        self.temp_importance = 0.0      # Used for on-the-fly importance calculation in
                                        # network.get_importance()
        self.name = name
        self.rf = []                    # list of owning RFBinding objects
        self.semantic = {}              # dict of owned Semantic objects in {ref:weight} format

    #-------------------------------------
    # Token-based input methods
    #-------------------------------------

    def get_bu_input(self):
        '''Returns a weighted sum of the bottom-up inputs connected to this unit.

        This is the overloaded method that belongs to OPs.  It gets inputs from the semantics
        connected to this unit.'''

        self.bu_input = sum(s.act * weight for s, weight in self.semantic.iteritems())
        
    def get_td_input(self):
        '''Returns a weighted sum of top-down inputs connected to this unit.

        This is the overloaded method that belongs to OPs.  It gets inputs from the rfs
        connected to this unit.'''

        self.td_input = sum(rf.act for rf in self.rf) / (1 + len(self.rf))

    def get_outprop_input(self, analog):
        '''Returns the activations of RFs that inhibit the current OP.  RFs do not inhibit their
        own OPs.'''

        self.outprop_input = sum(rf.act for rf in analog.rfbindings if rf not in self.rf)

    #--------------------------------
    # Net input methods
    #--------------------------------

    def get_driver_net_input(self):
        '''Returns the net input to this driver unit.

        This is the method that belongs to OPs.  It gets input solely from its td_input
        field.'''

        self.net_input = self.td_input

    def get_recipient_net_input(self):
        '''Returns the net input to this recipient unit.

        This is the method that belongs to OPs.  It does not include within_class_input.'''

        self.net_input = (self.bu_input + self.td_input) + self.total_hebb_input - self.outprop_input

class Obj(OP):
  def __init__(self, name):
      OP.__init__(self, name)

class Pred(OP):
  def __init__(self, name):
      OP.__init__(self, name)


class RFBinding(Token):
    def __init__(self, name):
        Token.__init__(self)
        self.name = name
        self.prop = None          # owning Parent Prop object
        self.obj = None           # owned Obj objects or Child Prop object
        self.pred = None          # owned Pred object
        self.fired = False        # for Top-Down OK

    #--------------------------------
    # Token-based input methods
    #--------------------------------

    def get_bu_input(self, action_flag):
        '''Returns a weighted sum of the bottom-up inputs connected to this unit.

        This is the overloaded method that belongs to RFs.  It gets inputs from the preds, objs,
        and child props connected to this unit.'''
                    
        if type(self.obj) is Prop and self.obj.mode == 'child':
            if action_flag == 'retrieve':
                obj_act = 0.0
            else:
                obj_act = self.obj.act
        elif type(self.obj) is Obj:
            obj_act = self.obj.act
        else:
            obj_act = 0.0

        try:
            pred_act = self.pred.act
        except AttributeError:
            pred_act = 0.0
        
        self.bu_input = obj_act + pred_act

    def get_td_input(self):
        '''Returns a weighted sum of the top-down inputs connected to this unit.

        This is the overloaded method that belongs to RFs.  It gets inputs from the parent props
        connected to this unit.'''

        try:
            if self.prop.mode == 'parent':
                self.td_input = self.prop.act
            else:
                self.td_input = 0.0
        except AttributeError:
            self.td_input = 0.0


    def get_within_class_input(self, analog):
        '''Returns a normalized sum of the activations of competing rfbinding
        units that inhibit the current RF.

        * * * Inhibition and infant units * * *

        When a unit is inferred, it needs to remain the most active
        unit of its type until it is fully connected ("grown up").
        Infants are at a disadvantage because, unlike fully-connected
        units, they don't receive excitatory top-down and bottom-up
        input. In order to insure that infants remain the most active
        units of their type, they should not be inhibited by other
        units.  The only case where we can know with certainly whether
        a unit is an infant is the RF, so infant RFs do not get
        within-class or out-prop inhibition.

        Infants should immediately inhibit other units, of course.'''
      
        rf_act = sum(rf.act for rf in analog.rfbindings if rf is not self)
        thresh_act = [i.act for i in analog.rfbindings if i.act > theta_i]

        if self.infant == False:
            self.within_class_input = rho_rf_i * rf_act /(1 + len(thresh_act))
        else:
            self.within_class_input = 0.0
                                                
    def get_outprop_input(self, analog):
        '''Returns a sum of the activations of parent props that inhibit the current RF.
        Parent props do not inhibit their own RFs.

        This method also allows other Episode groups to inhibit the
        current RF.  In principle, all units that belong to a group
        connect directly to that group; the constituent props are
        "flat" from the point of view of the group.  The group should
        be able to inhibit or support all of them equally (in
        practice, a lot of stuff in done through the props).  In this
        case, allowing *other* groups to inhibit this group's RFs is a
        proxy for allowing groups to pass around inhibitory signals
        that they send to all of their constituents.'''

        if self.infant == False:
            self.outprop_input = sum(prop.act for prop in analog.props
                                     if prop.mode == 'parent' and prop is not self.prop) + \
                                 rho_g_i * sum(group.act for group in analog.groups if type(group)
                                               is Episode and group not in self.prop.group)
##                                            sum(group.act for group in analog.groups if type(group)
##                                               is Effect and group not in self.prop.group))
        else:
            self.outprop_input = 0.0

class Prop(Token):
    def __init__(self, name):
        Token.__init__(self)
        self.name = name
        self.mode = 'neutral'       # 'child', 'parent', or 'neutral'
        self.old_mode = None
        self.priority = None
        self.parent_rf = []
        self.group = []             # list of owning Group objects
        self.episode_role = None    # 'state','action'
        self.rf = []                # list of owned RFBinding objects
        self.child_evidence = 0.0
        self.parent_evidence = 0.0
        self.importance = 1.0
        self.max_act = 0.0
        self.old_max_act = 0.0

        self.fired = False

        # These fields are currently not being used
        self.centrality = 0.0
        self.priority = 0.0         # priority = readiness * importance * support
        self.readiness = 1.0
        self.support = 0.0
        

    #-----------------------------------
    # Hebb-based input methods
    #-----------------------------------
 
    def get_mode_evidence(self, driver, hebbs):
        '''Returns evidence about the modes of props in the driver analog that the current unit
        is connected to.'''

        hebbset = self.get_hebbset(driver)
        c_hebb_act = []
        c_hebb_w = []
        p_hebb_act = []
        p_hebb_w = []
        
        for hebb in hebbset:
            driver_mode = hebbs[hebb].owners[driver].mode
            if driver_mode == 'child':
                c_hebb_act.append(hebbs[hebb].owners[driver].act)
                c_hebb_w.append(hebbs[hebb].perm_wt)
            if driver_mode == 'parent':
                p_hebb_act.append(hebbs[hebb].owners[driver].act)
                p_hebb_w.append(hebbs[hebb].perm_wt)

        self.child_evidence = sum(x*y for x,y in itertools.izip(c_hebb_act, c_hebb_w))
        self.parent_evidence = sum(x*y for x,y in itertools.izip(p_hebb_act, p_hebb_w))

    def setmode(self, phi):
        '''Sets the mode of the prop based on bottom-up, top-down, and hebb evidence.'''
        
        a = self.bu_input
        b = self.parent_evidence
        c = self.td_input
        d = self.child_evidence

        self.old_mode = self.mode

        if a + b - c - d > phi:
            self.mode = 'parent'
        elif a + b - c - d < -phi:
            self.mode = 'child'
        else:
            self.mode = 'neutral'

    #-----------------------------------
    # Token-based input methods
    #-----------------------------------

    def get_bu_input(self):
        '''Returns a weighted sum of bottom-up inputs connected to this unit.'''

        self.bu_input = sum(rf.act for rf in self.rf)

    def get_td_input(self):
        '''Returns a weighted sum of top-down inputs connected to this unit.'''

        self.td_input = sum(rf.act for rf in self.parent_rf)

    def get_within_class_input(self, analog):
        '''Returns a normalized sum of the activations of competing props that inhibit the
        current prop.'''    
        
        if self.mode == 'neutral' or self.mode == 'parent':      
            prop_act = sum(prop.act for prop in analog.props if prop is not self and prop.mode != 'child')
            thresh_act = [i.act for i in analog.props if i.act > theta_i and i.mode != 'child']
            self.within_class_input = prop_act /(1 + len(thresh_act))
        elif self.mode == 'child':
            self.within_class_input = 0.0

    def get_outprop_input(self, analog):
        '''Returns the activations of RFs that inhibit the current child prop.  RFs do not
        inhibit their own child props.

        This method will execute when all RFs are grown up; i.e., infant == False.
        If this works - using RFs as key to whether there are infants in the system -
        it should be extended to get_outprop_input and get_within_class input for all
        units.'''
  
        if self.mode == 'child' and self.parent_rf:
            self.outprop_input = sum(rf.act for rf in analog.rfbindings if rf not in self.parent_rf)
        elif self.mode == 'parent' and self.group:
            self.outprop_input = rho_g_p * sum(group.act for group in analog.groups if type(group)
                                               is Episode and group not in self.group)
##                                            sum(group.act for group in analog.groups if type(group)
##                                               is Effect and group not in self.group))
        else:
            self.outprop_input = 0.0
   
    #-----------------------------------
    # Retrieval methods
    #-----------------------------------

    def get_max_act(self):
        self.old_max_act = self.max_act
        if self.act > self.max_act:
            self.max_act = self.act            

class Group(Token):
    '''Groups collect the elements of rules (which are also the elements of causal relations).
    A complete rule/causal relation has three elements:
    1.  Causal episode (L-hand side):
        a. Episode group
        b. State propositional elements
        c. Action propositional elements
    2.  Effect episode (R-hand side):
        a. Episode group
        b. State propositional elements
    3.  Causal Link (collects Causal episode group and Effect episode group together):
        a. Causal Link group

    For programming simplicity, the Episode groups will only specify their
    constituent Props.  However, these groups are actually directly connected to
    ALL of their tokens.  The Prop information will be used to get all of that
    info back.

    Causal power is different from value in two respects:
    1. Causal power is scale-free.  It captures the likelihood of the effect
    coming to pass, and therefore ranges [-1:1].  Value asypmtotes towards the
    value of the reward.
    2. Causal power is proportional to |value|.  Negative value implies the
    power to generate bad outcomes.

    It would be ideal for +/- causal power to implicitly capture
    generative/preventative causes.  We'll try to get that behavior out of the
    system.

    With this in mind, it may be correct to capture value on the Causal episode,
    but causal power on the Causal Link.  This allows each cause to have
    multiple independent effects.'''

    
    def __init__(self, name):
        Token.__init__(self)
        self.name = name
        self.semantic = {}              # Episode groups have 'cause' or 'effect' semantics;
                                        # Link groups have 'link' semantics
        self.inhibited = False          # Temporarily prohibited for retrieval
        self.bu_input = 0.0
        self.td_input = 0.0
        self.within_class_input = 0.0
        self.outprop_input = 0.0

        # Non-propositional information
        self.similarity_strength = 0.0
        self.value = 0.0                # malleable hedonic value
        self.reward = 0.0               # fixed reward; mutually exclusive with value
        self.causal_power = 0.0
        self.act_memory = []

    def update_act_memory(self):
        self.act_memory.append(self.act)

    def set_fired(self, fired):
        pass

class Link(Group):
    def __init__(self, name):
        Group.__init__(self, name)
        self.episode = []

    def get_driver_bu_input(self):
        if self.phase_set_flag == True:
            self.bu_input = sum(i.act for i in self.episode)
        else:
            self.bu_input = 0.0

    def get_recip_bu_input(self, analog):
        group_input = sum(i.act for i in self.episode)            
        semantic_input = sum(s.act * weight for s, weight in self.semantic.iteritems())
        self.bu_input = group_input + semantic_input

    def get_outprop_input(self, analog):
        pass

    def get_within_class_input(self, analog):
        group_act = sum(group.act for group in analog.groups if group is not self
                        and type(group) is Link)
        thresh_act = [i.act for i in analog.groups if i.act > theta_i and
                      type(i) is Link]

        if self.infant == False:
            self.within_class_input = rho_g * group_act /(1 + len(thresh_act))
        else:
            self.within_class_input = 0.0

class Episode(Group):
    def __init__(self, name):
        Group.__init__(self, name)
        self.state = []                 # Episode groups have state props, and sometimes action props
        self.link = None                # Episode groups have parent links
        self.fired = False              # Have you been fired yet?

    def set_fired(self, fired):
        self.fired = fired

    def get_driver_bu_input(self):
        if self.phase_set_flag == True:
            self.bu_input = sum(prop.act for prop in self.state)
        else:
            self.bu_input = 0.0

    def get_recip_bu_input(self, analog):
        semantic_input = sum(s.act * weight for s, weight in self.semantic.iteritems())
        prop_input = sum(prop.act for prop in self.state)
        
        if semantic_input > theta_c:
            self.bu_input = prop_input + semantic_input
        else:
            self.bu_input = 0.0

    def get_within_class_input(self, analog):
        group_act = sum(group.act for group in analog.groups if group is not self
                        and type(group) == Episode)
        thresh_act = [i.act for i in analog.groups if i.act > theta_i and type(i) == Episode]

        if self.infant == False:
            self.within_class_input = rho_g * group_act /(1 + len(thresh_act))
        else:
            self.within_class_input = 0.0

    def get_outprop_input(self, analog):
        pass

    def get_similarity_strength(self, direction):
        
        props = [prop.max_act for prop in self.state] # if prop.episode_role == 'state']
        self.similarity_strength = sum(pow(i, 2) for i in props)/len(props) #pow(sum(props)/(1 + len(props)), 2)

#-----------------------------------
# Semantic class
#-----------------------------------
class Semantic(object):
    def __init__(self, name):
        self.name = name
        self.net_input = 0.0
        self.act = 0.0
        self.delta = 0.0
        self.op = {}            # dict of owning OP objects in <ref>:weight format

    def get_semantic_net_input(self, analogs):
        '''Returns the net input to this semantic from all OPs to which it's connected.

        Only invoke this in a recipient analog when top-down okay == True.'''
        
        op_act = []
        op_w = []
        for op, weight in self.op.iteritems():
            for analog in analogs.itervalues():
                if analog.analog_type == 'driver' or analog.analog_type == 'recipient' and analog.top_down_ok == True:
                    if op in analog.preds or op in analog.objs or op in analog.groups:
                        op_act.append(op.act)
                        op_w.append(weight)
        self.net_input = sum(x*y for x,y in itertools.izip(op_act, op_w))
       
    def get_semantic_act(self, semantics):
        '''Returns the activation of the semantic unit.  We add a
        small amount of random noise so that non-overlapping preds can
        map.  When no effect group exists in the target, the target
        uses the action to perceive the relevant part of the board.
        So we want to map put(disk A, peg 1) to on_top(disk A, peg 1).
        Since action preds share no semantics with state preds, the
        respective props don't map, and instead you end up with
        smaller(disk A, disk B).

        Note that it's a GOOD thing that actions can't map to states!
        If actions and states share semantics (or if the noise
        component is large), the organization of the effect group
        overwhelms the preferred mappings.  Imagine you have the
        following groupings:

        Target:             on_top(disk A, peg 1)
        Source:
            cause state:    on_top(disk X, peg I)
            cause action:   put (disk X, peg J)
            effect:         on_top (disk X, peg J)
        You want to map A to X and 1 to I.  But the weight of the
        evidence is for A to X and 1 to J, which results in all kinds
        of weird out-mappings.  The only way to prevent this is to
        give precedence to structure, and keep the action preds from
        throwing their weight behind a mapping that reflects the
        future state of the world, not the present state.

        In conclusion: It is necessary that action preds map to state
        preds during Effect group creation, but it is disastrous at
        any other time.'''

        max_net_input = max(abs(i.net_input) for i in semantics)
        noise = 0
        #noise = random.uniform(-0.01, 0.01)
        self.act = self.net_input/max(max_net_input, 1) + noise

#----------------------------------------------
# Analog and other control structures classes
#----------------------------------------------
    
class Analog(object):
    def __init__(self, name):
        self.name = name
        self.sequences = []
        self.groups = []
        self.props = []
        self.rfbindings = []
        self.preds = []
        self.objs = []

        # Used by symparser
        self.role_correspondence = {}
        self.obj_correspondence = {}

        # Most active units and their hebbs
        self.driver_pred_act = 0.0
        self.driver_obj_act = 0.0
        self.driver_rf_act = 0.0
        self.driver_prop_act = 0.0
        self.driver_c_prop_act = 0.0
        self.driver_link_act = 0.0
        self.driver_eq_act = 0.0
        self.driver_episode_act = 0.0
##        self.driver_cause_act = 0.0
##        self.driver_effect_act = 0.0
        self.driver_pred_hebb = 0.0
        self.driver_obj_hebb = 0.0
        self.driver_rf_hebb = 0.0
        self.driver_prop_hebb = 0.0
        self.driver_c_prop_hebb = 0.0
        self.driver_link_hebb = 0.0
        self.driver_eq_hebb = 0.0
        self.driver_episode_hebb = 0.0
##        self.driver_cause_hebb = 0.0
##        self.driver_effect_hebb = 0.0
        self.driver_pred = None
        self.driver_obj = None
        self.driver_rf = None
        self.driver_prop = None
        self.driver_c_prop = None
        self.driver_link = None
        self.driver_eq = None
        self.driver_episode = None
##        self.driver_cause = None
##        self.driver_effect = None

        # Control switches
        self.top_down_ok = False            # All RFs in driver have fired once -> True
        self.analog_type = None             # 'driver','recipient'
        self.ready_to_learn = False
        
        # Currently not being used
        self.analog_location = None         # 'target','source','world','ltm','schema'
        self.learning_permitted = True      # unsupervised learning allowed
        self.mappings_required = True       # unsupervised learning requires adequate mappings
                                            # to license learning

    def update_global_inhibitor(self, flag='minor'):
        '''Resets activation of all symbols in all units to 0.'''

        self.driver_pred_act = 0.0
        self.driver_obj_act = 0.0
        self.driver_rf_act = 0.0
        self.driver_prop_act = 0.0
        self.driver_c_prop_act = 0.0
        self.driver_link_act = 0.0
        self.driver_eq_act = 0.0
        self.driver_episode_act = 0.0
##        self.driver_cause_act = 0.0
##        self.driver_effect_act = 0.0
        self.driver_pred_hebb = 0.0
        self.driver_obj_hebb = 0.0
        self.driver_rf_hebb = 0.0
        self.driver_prop_hebb = 0.0
        self.driver_c_prop_hebb = 0.0
        self.driver_link_hebb = 0.0
        self.driver_eq_hebb = 0.0
        self.driver_episode_hebb = 0.0
##        self.driver_cause_hebb = 0.0
##        self.driver_effect_hebb = 0.0
        self.driver_pred = None
        self.driver_obj = None
        self.driver_rf = None
        self.driver_prop = None
        self.driver_c_prop = None
        self.driver_link = None
        self.driver_eq = None
        self.driver_episode = None
##        self.driver_cause = None
##        self.driver_effect = None

        for value in self.preds:
            value.bu_input = 0.0 
            value.td_input = 0.0 
            value.within_class_input = 0.0
            value.outprop_input = 0.0
            value.hebb_inhib = 0.0
            value.hebb_excite = 0.0
            value.max_hebb_w = 0.0
            value.second_max_hebb_w = 0.0
            value.total_hebb_input = 0.0
            value.net_input = 0.0
            value.act_delta = 0.0
            value.act = 0.0
            value.old_act = 0.0
        for value in self.objs:
            value.bu_input = 0.0 
            value.td_input = 0.0 
            value.within_class_input = 0.0
            value.outprop_input = 0.0
            value.hebb_inhib = 0.0
            value.hebb_excite = 0.0
            value.max_hebb_w = 0.0
            value.second_max_hebb_w = 0.0
            value.total_hebb_input = 0.0
            value.net_input = 0.0
            value.act_delta = 0.0
            value.act = 0.0
            value.old_act = 0.0
        for value in self.rfbindings:
            value.bu_input = 0.0 
            value.td_input = 0.0 
            value.within_class_input = 0.0
            value.outprop_input = 0.0
            value.hebb_inhib = 0.0
            value.hebb_excite = 0.0
            value.max_hebb_w = 0.0
            value.second_max_hebb_w = 0.0
            value.total_hebb_input = 0.0
            value.net_input = 0.0
            value.act_delta = 0.0
            value.act = 0.0
            value.old_act = 0.0
        for value in self.props:
            if value.mode == 'child':
                value.bu_input = 0.0 
                value.td_input = 0.0 
                value.within_class_input = 0.0
                value.outprop_input = 0.0
                value.hebb_inhib = 0.0
                value.hebb_excite = 0.0
                value.max_hebb_w = 0.0
                value.second_max_hebb_w = 0.0
                value.total_hebb_input = 0.0
                value.net_input = 0.0
                value.act_delta = 0.0
                value.act = 0.0
                value.old_act = 0.0
                value.child_evidence = 0.0
                value.parent_evidence = 0.0
        if flag == 'major':
            for value in self.preds:
                value.retrieved = False
            for value in self.objs:
                value.retrieved = False
            for value in self.rfbindings:
                value.retrieved = False
            for value in self.props:
                value.retrieved = False
                value.bu_input = 0.0 
                value.td_input = 0.0 
                value.within_class_input = 0.0
                value.outprop_input = 0.0
                value.hebb_inhib = 0.0
                value.hebb_excite = 0.0
                value.max_hebb_w = 0.0
                value.second_max_hebb_w = 0.0
                value.total_hebb_input = 0.0
                value.net_input = 0.0
                value.act_delta = 0.0
                value.act = 0.0
                value.old_act = 0.0
                value.child_evidence = 0.0
                value.parent_evidence = 0.0
            for value in self.groups:
                value.retrieved = False
                value.bu_input = 0.0 
                value.td_input = 0.0 
                value.within_class_input = 0.0
                value.outprop_input = 0.0
                value.hebb_inhib = 0.0
                value.hebb_excite = 0.0
                value.max_hebb_w = 0.0
                value.second_max_hebb_w = 0.0
                value.total_hebb_input = 0.0
                value.net_input = 0.0
                value.act_delta = 0.0
                value.act = 0.0
                value.old_act = 0.0

        
