'''symparser.py

The symparser module creates a working memory network from a sym file.  A sym file
always contains a Symfile() class with a self.sym field describing the network
being created.

The code below creates a gargantuan Symparser object that really only serves as
a namespace (i.e. a Singlton).  It stores all of the objects that make up the
network.'''


#-----------------------------------------
# Imports and setup
#-----------------------------------------
import copy, random, sys
import network as nw
from perstypes import HebbDict, Episode, Link, Prop, Obj, Pred, RFBinding, Sequence
from persparms import e_lambda


class Symparser(object):
    def __init__(self):

        self.sym = None
        self.world = None           # This will contain the portion of the world that is to
                                    # be predicated this turn.  This means ALL units.
        self.prop_queue = []        # Current firing queue for driver
        self.group_queue = []
        
        self.pred_semantics = []
        self.obj_semantics = []
        self.episode_semantics = []
        self.link_semantics = []
        self.analogs = {}
        self.hebbs = HebbDict()

    def print_analog(self, analog_name, group_flag=False, prop_flag=True, concise_groups=False):
        if group_flag == True:
            print analog_name.capitalize(), 'groups'
            if concise_groups == True:
                for group in self.analogs[analog_name].groups:
                    if type(group) is Episode:
                        #if any(i for i in group.semantic.iterkeys() if i.name == 'cause'):
                        print ' ', group.name, ':', [i.name for i in group.state], group.link
            else:
                group_print_list = []
                for group in self.analogs[analog_name].groups:
                    if type(group) is Link:
                        cause = None
                        effect = None
                        cause_name = None
                        effect_name = None
                        for episode in group.episode:
                            cause_s = [i for i in episode.semantic.iterkeys() if i.name == 'cause']
                            effect_s = [i for i in episode.semantic.iterkeys() if i.name == 'effect']
                            if any(cause_s) and not any(effect_s):
                                cause = episode
                            elif not any(cause_s) and any(effect_s):
                                effect = episode
                        if cause is not None:
                            cause_name = cause.name
                        if effect is not None:
                            effect_name = effect.name
                        residual = [i for i in group.episode if i != cause and i != effect]
                        group_print_list.append([group.name, cause_name, effect_name, residual])
                group_print_list.sort()
                for group_name, group_cause, group_effect, residual in group_print_list:
                    print ' ', group_name, ':', group_cause, group_effect, [i.name for i in residual]
                    
                group_print_list = []
                for group in self.analogs[analog_name].groups:
                    if type(group) is Episode: # or type(group) is Effect:
                        if group.link is not None:
                            group_print_list.append([group.name, [i.name for i in group.state], group.link.name])
                        else:
                            group_print_list.append([group.name, [i.name for i in group.state], group.link])
                group_print_list.sort()
                for group_name, group_state, group_link, in group_print_list:
                    print ' ', group_name, ':', group_state, group_link
                    
##                group_print_list = []
##                for group in self.analogs[analog_name].groups:
##                    if type(group) is Equivalence:
##                        group_print_list.append([group.name, [i.name for i in group.episode]])
##                group_print_list.sort()
##                for group_name, group_set in group_print_list:
##                    print ' ', group_name, ':', group_set
                    
        if prop_flag == True:
            print analog_name.capitalize(), 'props'
            prop_print_list = []
            for prop in self.analogs[analog_name].props:
                preds = []
                for rf in prop.rf:
                    place = []
                    if rf.pred is not None:
                        place.append(rf.pred.name)
                    if rf.obj is not None:
                        place.append(rf.obj.name)
                    preds.append(place)
                preds.sort()
                prop_print_list.append([prop.name, preds, [i.name for i in prop.group]])
            prop_print_list.sort()
            for prop_name, pred_set, group_set in prop_print_list:
                print ' ', prop_name, pred_set, group_set
##        print '  Props', [i.name for i in self.analogs[analog_name].props]
##        print '  RFs', [i.name for i in self.analogs[analog_name].rfbindings]
##        print '  Preds', [i.name for i in self.analogs[analog_name].preds]
##        print '  Objs', [i.name for i in self.analogs[analog_name].objs]


    def print_hebbs(self):

        type_list = [Link, Episode, Prop, RFBinding, Pred, Obj]
        print
        print 'Hebbs'

        for unit_type in type_list:
            print ' ', str(unit_type)
            unit_hebb_list = []
            for h_key, h in self.hebbs.iteritems():
                if h.perm_wt > 0.01 and type(h.owners.values()[0]) is unit_type:
                    h_list = [(i, j.name) for i,j in h.owners.iteritems()]
                    h_list.sort()
                    unit_hebb_list.append([[i[1] for i in h_list], h.perm_wt])
            unit_hebb_list.sort()
            for units, weight in unit_hebb_list:
                print '  ', units, weight

    #-----------------------------------------------
    # Firing sequence methods
    #-----------------------------------------------

    def set_mappable(self, direction, sequence, driver=None, action_flag=None,
                     reset_only=True, to_analog=None):
        
        if reset_only == True:
            for unit_list in sequence.__dict__.itervalues():
                try:
                    for unit in unit_list:
                        try:
                            unit.set_mappable(True)
                        except AttributeError:
                            pass
                except TypeError:
                    pass

        # During mapping and inference:
        #  1. When the target becomes the driver:
        #    A. Only the elements of source_seq.ordered_groups[-1] should be mappable
        #  2. When the source becomes the driver:
        #    A. All source units become mappable
        #    B. Only target units with hebbs AND mappable == True are mappable; all others False

        # This assumes that you have already called set_mappable() with reset_only=True on the source sequence
        elif driver == 'source' and action_flag == 'infer':
            if direction == 'explain':
                
                # Limit the Target unit to those that are eligible from the mapping phase
                for unit_list in sequence.__dict__.itervalues():
                    try:
                        for unit in unit_list:
                            if to_analog in unit.hebbsets and any(unit.hebbsets[to_analog]) and unit.mappable == True:
                                unit.set_mappable(True)
                            else:
                                unit.set_mappable(False)
                    except TypeError:
                        pass

        # Set mappable for SOURCE sequence        
        elif driver == 'target' and action_flag == 'map':
            if direction == 'explain':
                j = len(sequence.ordered_groups)
                if any(i for i in sequence.ordered_groups[-1].semantic.iterkeys() if i.name == 'effect'):
                    # Set all preceding group elements to False
                    for i in xrange(0,j-1):
                        sequence.ordered_groups[i].set_mappable(False)
                        try:
                            for prop in sequence.ordered_groups[i].state:
                                prop.set_mappable(False)
                                for rf in prop.rf:
                                    rf.set_mappable(False)
                                    rf.pred.set_mappable(False)
                                    rf.obj.set_mappable(False)
                        except AttributeError:
                            pass
                    # Reset latest effect group elements to True
                    sequence.ordered_groups[-1].set_mappable(False)
                    try:
                        for prop in sequence.ordered_groups[-1].state:
                            prop.set_mappable(True)
                            for rf in prop.rf:
                                rf.set_mappable(True)
                                rf.pred.set_mappable(True)
                                if type(rf.obj) == Obj:
                                    rf.obj.set_mappable(True)
                    except AttributeError:
                        pass
                else:
                    print 'No groups, or last group in sequence is not an effect.  source:', __name___ , '.set_mappable()'
                        

        # This implements mappable for the Target driver prior to retrieval
        elif action_flag == 'retrieve':
            if any(sequence.ordered_groups):
                for group in sequence.ordered_groups:
                    if any(group.hebbsets[to_analog]):
                        group.set_mappable(False)
                        
                # ordered_groups = [C.1, L.1, E.1, C.2, L.2, E.2....];
                # For explain, we want the earliest cause where mappable = False
                # If we assume groups inserted in correct order, the earliest group will be a cause             
                j = len(sequence.ordered_groups)
                
                if direction == 'explain':
                    #if type(sequence.ordered_groups[0]) is Cause:
                    if any(i for i in sequence.ordered_groups[0].semantic.iterkeys() if i.name == 'cause'):
                        # Set all subsequent group elements to False
                        for i in xrange(1,j):
                            try:
                                for prop in sequence.ordered_groups[i].state:
                                    if any(prop.hebbsets[to_analog]):
                                        prop.set_mappable(False)
                                    for rf in prop.rf:
                                        if any(rf.hebbsets[to_analog]):
                                            rf.set_mappable(False)
                                        if any(rf.pred.hebbsets[to_analog]):
                                            rf.pred.set_mappable(False)
                                        if any(rf.obj.hebbsets[to_analog]):
                                            rf.obj.set_mappable(False)
                            except AttributeError:
                                pass
                        # Reset earliest cause group elements to True
                        try:
                            for prop in sequence.ordered_groups[0].state:
                                prop.set_mappable(True)
                                for rf in prop.rf:
                                    rf.set_mappable(True)
                                    rf.pred.set_mappable(True)
                                    if type(rf.obj) == Obj:
                                        rf.obj.set_mappable(True)
                        except AttributeError:
                            pass
                    else:
                        print 'No groups, or first group in sequence is not a cause.  source:', __name___ , '.set_mappable()'

                if direction == 'predict':
                    #if type(sequence.ordered_groups[-1]) is Effect:
                    if any(i for i in sequence.ordered_groups[-1].semantic.iterkeys() if i.name == 'effect'):
                        # Set all subsequent group elements to False
                        for i in xrange(0,j-1):
                            try:
                                for prop in sequence.ordered_groups[i].state:
                                    if any(prop.hebbsets[to_analog]):
                                        prop.set_mappable(False)
                                    for rf in prop.rf:
                                        if any(rf.hebbsets[to_analog]):
                                            rf.set_mappable(False)
                                        if any(rf.pred.hebbsets[to_analog]):
                                            rf.pred.set_mappable(False)
                                        if any(rf.obj.hebbsets[to_analog]):
                                            rf.obj.set_mappable(False)
                            except AttributeError:
                                pass
                        # Reset earliest cause group elements to True
                        try:
                            for prop in sequence.ordered_groups[-1].state:
                                prop.set_mappable(True)
                                for rf in prop.rf:
                                    rf.set_mappable(True)
                                    rf.pred.set_mappable(True)
                                    if type(rf.obj) == Obj:
                                        rf.obj.set_mappable(True)
                        except AttributeError:
                            pass
                    else:
                        print 'No groups, or last group in sequence is not an effect.  source:', __name___ , '.set_mappable()'


    def set_group_firing_sequence(self, sequence, direction, action_flag):
        if action_flag == 'retrieve':
            self.group_queue = [None]
        else:
            if direction == 'predict':
                if any(sequence.ordered_groups):
                    for group in sequence.ordered_groups:
                        if hasattr(group, 'state') and group.fired == False:
                            #if any([i.mappable for i in group.state]):
                            self.group_queue.append(group)
                else:
                    self.group_queue = [None]
            elif direction == 'explain':
                if any(sequence.ordered_groups):
                    j = len(sequence.ordered_groups)
                    while j != 0:
                        j -= 1
                        if hasattr(sequence.ordered_groups[j], 'state') and sequence.ordered_groups[j].fired == False:
                            #if any([i.mappable for i in sequence.ordered_groups[j].state]):
                            self.group_queue.append(sequence.ordered_groups[j])
                else:
                    self.group_queue = [None]

    def set_prop_firing_sequence(self, driver, firing_episode, action_flag):            
        if firing_episode is None:
            self.prop_queue = [prop for prop in self.analogs[driver].props if prop.mappable == True]
        else:
            self.prop_queue = [prop for prop in firing_episode.state]   # if prop.mappable == True]


    #-----------------------------------------------
    # Object creation methods
    #-----------------------------------------------

    def create_analogs(self):
        for analog_name in self.sym.sym.iterkeys():
            analog = nw.create_analog(analog_name, self.analogs)
            if analog_name == 'schema':
                analog.ready_to_learn = True

    # All of the following methods are invoked once for each analog_dict.  For building the
    # original network, the syntax is:
    #   for analog_name, analog_dict in self.sym.sym.iteritems():
    # This decouples the methods from the symfile, which allows them to be used for building
    # world props (e.g.).

    def create_objs(self, analog_name, sequence_text, sequence):
        for obj_name, semantics in sequence_text['obj'].iteritems():
            if obj_name not in [i.name for i in sequence.objs]:
                unit_semantics = []
                for semantic_name in semantics:
                    nw.create_semantic(semantic_name, unit_semantics, self.obj_semantics)
                obj = nw.create_op('obj', self.analogs[analog_name], obj_name, sequence,
                                   unit_semantics)

    def create_groups(self, analog_name, sequence_text, sequence):

        for ordered_group in sequence_text['group']:
            for group_set_name, group_set in ordered_group.iteritems():

                cause = None
                if 'cause' in group_set:
                    cause_name = group_set_name + '.C'
                    group_semantics = []
                    nw.create_semantic(cause_name, group_semantics, self.episode_semantics)
                    nw.create_semantic('cause', group_semantics, self.episode_semantics)
                    cause = nw.create_group(self.analogs[analog_name], cause_name, sequence, 'episode',
                                            group_semantics, value = group_set['cause']['value'],
                                            reward = group_set['cause']['reward'])
                    if 'prop' in group_set['cause']:
                        for prop_name in group_set['cause']['prop']:
                            prop = [i for i in sequence.props if i.name == prop_name].pop()
                            prop.group.append(cause)
                            cause.state.append(prop)

                link = None
                if 'link' in group_set:
                    link_name = group_set_name + '.L'
                    group_semantics = []
                    nw.create_semantic(link_name, group_semantics, self.link_semantics)
                    nw.create_semantic('link', group_semantics, self.link_semantics)
                    link = nw.create_group(self.analogs[analog_name], link_name, sequence, 'link',
                                           group_semantics, causal_power = group_set['link']['causal_power'])
                    if cause is not None:
                        cause.link = link
                        link.episode.append(cause)
                    
                if 'effect' in group_set:
                    effect_name = group_set_name + '.E'
                    group_semantics = []
                    nw.create_semantic(effect_name, group_semantics, self.episode_semantics)
                    nw.create_semantic('effect', group_semantics, self.episode_semantics)
                    effect = nw.create_group(self.analogs[analog_name], effect_name, sequence, 'episode',
                                             group_semantics, value = group_set['effect']['value'],
                                             reward = group_set['effect']['reward'])                
                    if link is not None:
                        effect.link = link      # .append(link)
                        link.episode.append(effect) 
                        
                    if 'prop' in group_set['effect']:
                        for prop_name in group_set['effect']['prop']:
                            prop = [i for i in sequence.props if i.name == prop_name].pop()
                            prop.group.append(effect)
                            effect.state.append(prop)

    def create_preds(self, analog_name, sequence_text, sequence):
        #print sequence_text
        for pred_name, semantics in sequence_text['pred'].iteritems():
            if pred_name not in [i.name for i in sequence.preds]:
                role_count = semantics[0]
                        
                # Autocode preds
                for i in xrange(role_count):
                    unit_semantics = []
                    for semantic_name in semantics[1:]:
                        nw.create_semantic(semantic_name, unit_semantics, self.pred_semantics)
                    nw.create_semantic(pred_name + str(i+1), unit_semantics, self.pred_semantics)
                    pred = nw.create_op('pred', self.analogs[analog_name], pred_name + str(i+1),
                                        sequence, unit_semantics)

    def create_props(self, analog_name, sequence_text, sequence):
        for prop_name, content in sequence_text['prop'].iteritems():
        
            prop = nw.create_prop(self.analogs[analog_name], prop_name, sequence)
            
            if type(content[0]) is int:
                prop.priority = content.pop(0)
            
            pred_name = content.pop(0)
            pred_units = [(i.name, i) for i in sequence.preds if pred_name in i.name]
            pred_units.sort()
            pred_units = [i[1] for i in pred_units]

            for pred in pred_units:

                # Set action and state props
                if any(i for i in pred.semantic.iterkeys() if i.name == 'action'):
                    prop.episode_role = 'action'
                else:
                    prop.episode_role = 'state'

                # Connect prop elements with RFs
                try:
                    obj_name = content.pop(0)
                    if obj_name in [i.name for i in sequence.objs]:
                        obj = [i for i in sequence.objs if i.name == obj_name].pop()
                        nw.create_rf(self.analogs[analog_name], prop_name + '.' + pred.name[-1],
                                     sequence, prop, pred, obj)
                    elif obj_name in [i.name for i in sequence.props]:
                        obj = [i for i in sequence.props if i.name == obj_name].pop()
                        nw.create_rf(self.analogs[analog_name], prop_name + '.' + pred.name[-1],
                                     sequence, prop, pred, obj)
                except IndexError:
                    print analog_name, prop_name, pred.name

    #-----------------------------------------------
    # Network creation methods
    #-----------------------------------------------

    def parse_symfile(self):
        #if self.sym is not None:
        self.create_analogs()

        for analog_name, analog_list in self.sym.sym.iteritems():
            for sequence_text in analog_list:
                sequence = Sequence()
                self.analogs[analog_name].sequences.append(sequence)

                if 'obj' in sequence_text:
                    self.create_objs(analog_name, sequence_text, sequence)
                if 'pred' in sequence_text:
                    self.create_preds(analog_name, sequence_text, sequence)
                if 'prop' in sequence_text:
                    self.create_props(analog_name, sequence_text, sequence)
                if 'group' in sequence_text:
                    self.create_groups(analog_name, sequence_text, sequence)

    #-----------------------------------------------
    # Network management methods
    #-----------------------------------------------            

    def clear_hebbs(self):
        self.hebbs.clear()
        self.hebbs.hebb_count = 0
        for analog in self.analogs.itervalues():
            for group in analog.groups:
                group.hebbsets.clear()
            for prop in analog.props:
                prop.hebbsets.clear()
            for rf in analog.rfbindings:
                rf.hebbsets.clear()
            for pred in analog.preds:
                pred.hebbsets.clear()
            for obj in analog.objs:
                obj.hebbsets.clear()

    def clear_analog(self, analog_name, clear_flag='all'):
        if clear_flag == 'proposition':
            # Only used in world analog
            for group in self.analogs[analog_name].groups:
                for semantic in self.episode_semantics:
                    if group in semantic.op:
                        del semantic.op[group]
                for semantic in self.link_semantics:
                    if group in semantic.op:
                        del semantic.op[group]
                        
            self.analogs[analog_name].rfbindings = []
            self.analogs[analog_name].props = []
            self.analogs[analog_name].groups = []
            
            for pred in self.analogs[analog_name].preds:
                pred.rf = []
            for obj in self.analogs[analog_name].objs:
                obj.rf = []
                
        elif clear_flag == 'all':
            # Clear references to OPs in semantics
            for pred in self.analogs[analog_name].preds:
                for semantic in self.pred_semantics:
                    if pred in semantic.op:
                        del semantic.op[pred]
            for obj in self.analogs[analog_name].objs:
                for semantic in self.obj_semantics:
                    if obj in semantic.op:
                        del semantic.op[obj]
            for group in self.analogs[analog_name].groups:
                for semantic in self.episode_semantics:
                    if group in semantic.op:
                        del semantic.op[group]
                for semantic in self.link_semantics:
                    if group in semantic.op:
                        del semantic.op[group]
            self.analogs[analog_name].preds = []
            self.analogs[analog_name].objs = []
            self.analogs[analog_name].rfbindings = []
            self.analogs[analog_name].props = []
            self.analogs[analog_name].groups = []


