'''update.py            22 March 2006

Update methods for the model.'''
import perstypes

def set_phase_set(phase_set, firing_episode):
    '''Sets phase_set_flag = True for units in current phase set.'''

    if firing_episode is not None:
        firing_episode.phase_set_flag = True
        if firing_episode.link is not None:
            firing_episode.link.phase_set_flag = True

    if type(phase_set) == list:
        for prop in phase_set:
            prop.phase_set_flag = True
            for rf in prop.rf:
                rf.phase_set_flag = True
                rf.pred.phase_set_flag = True
                rf.obj.phase_set_flag = True
    else:
        phase_set.phase_set_flag = True
        for rf in phase_set.rf:
            rf.phase_set_flag = True
            rf.pred.phase_set_flag = True
            rf.obj.phase_set_flag = True

def initialize_props(analogs):
    '''Resets prop mode = 'neutral'.'''
    
    for analog in analogs.itervalues():
        for prop in analog.props:
            prop.mode = 'neutral'
       
def set_top_down_ok(analogs, top_down_flag=False):
    for analog in analogs.itervalues():
        analog.top_down_ok = top_down_flag

def get_hebb_buffers(hebbs):
    '''Updates buffers of hebbs connected to each recipient analog.'''

    for hebb in hebbs.itervalues():
        hebb.get_buffer_delta()
        hebb.update_buffer()
     
def get_hebb_div_max(hebbs):
    '''Gets the max hebb for divisive normalization.'''

    for hebb in hebbs.itervalues():
        hebb.get_max_buffer(hebbs)
                                    
def apply_hebb_div_max(hebbs):
    for hebb in hebbs.itervalues():
        hebb.apply_max_buffer()

def get_hebb_sub_max(hebbs):
    '''Gets the 2nd max hebb for subtractive normalization.'''
    
    for hebb in hebbs.itervalues():
        hebb.get_second_max_buffer(hebbs)

def settle_hebbs(hebbs, rho_he, rho_hi, gamma_h, delta_h, theta_h):
    '''Implements competetive activation between hebbs.'''

    counter = 1
    all_settled_flag = False

    while all_settled_flag == False and counter < 1000:
        all_settled_flag = True
        for hebb in hebbs.itervalues():
            hebb.get_second_max_act(hebbs)
            hebb.get_consistent_act(hebbs)
            hebb.get_total_input(rho_he, rho_hi)
            
        for hebb in hebbs.itervalues():
            settled_flag = hebb.update_activation(gamma_h, delta_h, theta_h)
            if settled_flag == False:
                all_settled_flag = False

        counter += 1

def update_hebb_weights(hebbs, eta):
    for hebb in hebbs.itervalues():
        hebb.get_w_delta(eta)
        hebb.update_perm_wt()

def flush_hebbs(hebbs):
    '''Initializes hebb units.'''
    
    for hebb in hebbs.itervalues():
        hebb.flush()

def get_driver_max_hebbs(analogs, driver, hebbs, theta_c, j):
    '''Gets the activation, name, and max hebb weight of the most active unit of each type from
    the driver to each recipient.  "driver_*" is the same for every
    recipient, because they belong to the most active unit in the driver.  "driver_*_hebb" is
    different for every recipient, because it's the max hebb from that driver unit to THIS
    recipient.

    The "driver_*_hebb" weight retrieved by this function is used by "nw.infer_new_units()",
    which uses it to name a newly-inferred unit in the recipient and decide which driver unit
    it should be mapped to.  The max hebb weight for an individual Token unit can be found by
    using that unit's "get_max_hebb_w()" function.'''

    # Get most active unit in driver
    for pred in analogs[driver].preds:
        if pred.phase_set_flag == True:
            if analogs[driver].driver_pred is not None:
                if pred.act > theta_c and pred.act > analogs[driver].driver_pred_act:
                    analogs[driver].driver_pred = pred
                    analogs[driver].driver_pred_act = pred.act
            else:
                if pred.act > theta_c:
                    analogs[driver].driver_pred = pred
                    analogs[driver].driver_pred_act = pred.act
    for obj in analogs[driver].objs:
        if obj.phase_set_flag == True:
            if analogs[driver].driver_obj is not None:
                if obj.act > theta_c and obj.act > analogs[driver].driver_obj_act:
                    analogs[driver].driver_obj = obj
                    analogs[driver].driver_obj_act = obj.act
            else:
                if obj.act > theta_c:
                    analogs[driver].driver_obj = obj
                    analogs[driver].driver_obj_act = obj.act
    for rf in analogs[driver].rfbindings:
        if rf.phase_set_flag == True:
            if analogs[driver].driver_rf is not None:
                if rf.act > theta_c and rf.act > analogs[driver].driver_rf_act:
                    analogs[driver].driver_rf = rf
                    analogs[driver].driver_rf_act = rf.act
            else:
                if rf.act > theta_c:
                    analogs[driver].driver_rf = rf
                    analogs[driver].driver_rf_act = rf.act
    for prop in analogs[driver].props:
        if prop.phase_set_flag == True and prop.mode == 'parent':
            if analogs[driver].driver_prop is not None:
                if prop.act > theta_c and prop.act > analogs[driver].driver_prop_act:
                    analogs[driver].driver_prop = prop
                    analogs[driver].driver_prop_act = prop.act
            else:
                if prop.act > theta_c:
                    analogs[driver].driver_prop = prop
                    analogs[driver].driver_prop_act = prop.act
        elif prop.phase_set_flag == True and prop.mode == 'child':
            if analogs[driver].driver_c_prop is not None:
                if prop.act > theta_c and prop.act > analogs[driver].driver_c_prop_act:
                    analogs[driver].driver_c_prop = prop
                    analogs[driver].driver_c_prop_act = prop.act
            else:
                if prop.act > theta_c:
                    analogs[driver].driver_c_prop = prop
                    analogs[driver].driver_c_prop_act = prop.act

    for group in analogs[driver].groups:
        if group.phase_set_flag == True:
            if type(group) == perstypes.Link:
                if analogs[driver].driver_link is not None:
                    if group.act > theta_c and group.act > analogs[driver].driver_link_act:
                        analogs[driver].driver_link = group
                        analogs[driver].driver_link_act = group.act
                else:
                    if group.act > theta_c:
                        analogs[driver].driver_link = group
                        analogs[driver].driver_link_act = group.act
            elif type(group) == perstypes.Episode: # and any(i for i in group.semantic.iterkeys() if i.name == 'cause'):
                if analogs[driver].driver_episode is not None:
                    if group.act > theta_c and group.act > analogs[driver].driver_episode_act:
                        analogs[driver].driver_episode = group
                        analogs[driver].driver_episode_act = group.act
                else:
                    if group.act > theta_c:
                        analogs[driver].driver_episode = group
                        analogs[driver].driver_episode_act = group.act

    # Get max hebb to each recipient analog
    for analog_name, analog in analogs.iteritems():
        if analog.analog_type == 'recipient':

            # Max pred hebb to current recipient
            if analogs[driver].driver_pred is not None:
                analog.driver_pred = analogs[driver].driver_pred
                analog.driver_pred_act = analogs[driver].driver_pred_act
                hebbset = analogs[driver].driver_pred.get_hebbset(analog_name)
                max_hebb = 0.0
                if len(hebbset) > 0:
                    max_hebb = max(hebbs[i].perm_wt for i in hebbset)
                analog.driver_pred_hebb = max_hebb
            
            # Max obj hebb to current recipient
            if analogs[driver].driver_obj is not None:
                analog.driver_obj = analogs[driver].driver_obj
                analog.driver_obj_act = analogs[driver].driver_obj_act
                hebbset = analogs[driver].driver_obj.get_hebbset(analog_name)
                max_hebb = 0.0
                if len(hebbset) > 0:
                    max_hebb = max(hebbs[i].perm_wt for i in hebbset)
                analog.driver_obj_hebb = max_hebb
            
            # Max RF hebb to current recipient
            if analogs[driver].driver_rf is not None:
                analog.driver_rf = analogs[driver].driver_rf
                analog.driver_rf_act = analogs[driver].driver_rf_act
                hebbset = analogs[driver].driver_rf.get_hebbset(analog_name)
                max_hebb = 0.0
                if len(hebbset) > 0:
                    max_hebb = max(hebbs[i].perm_wt for i in hebbset)
                analog.driver_rf_hebb = max_hebb
                
            # Max parent prop hebb to current recipient
            if analogs[driver].driver_prop is not None:
                analog.driver_prop = analogs[driver].driver_prop
                analog.driver_prop_act = analogs[driver].driver_prop_act
                hebbset = analogs[driver].driver_prop.get_hebbset(analog_name)
                max_hebb = 0.0
                if len(hebbset) > 0:
                    max_hebb = max(hebbs[i].perm_wt for i in hebbset)
                analog.driver_prop_hebb = max_hebb
            
            # Max child prop hebb to current recipient
            if analogs[driver].driver_c_prop is not None:
                analog.driver_c_prop = analogs[driver].driver_c_prop
                analog.driver_c_prop_act = analogs[driver].driver_c_prop_act
                hebbset = analogs[driver].driver_c_prop.get_hebbset(analog_name)
                max_hebb = 0.0
                if len(hebbset) > 0:
                    max_hebb = max(hebbs[i].perm_wt for i in hebbset)
                analog.driver_c_prop_hebb = max_hebb

            # Max causal link group hebb to current recipient
            if analogs[driver].driver_link is not None:
                analog.driver_link = analogs[driver].driver_link
                analog.driver_link_act = analogs[driver].driver_link_act
                hebbset = analogs[driver].driver_link.get_hebbset(analog_name)
                max_hebb = 0.0
                if len(hebbset) > 0:
                    max_hebb = max(hebbs[i].perm_wt for i in hebbset)
                analog.driver_link_hebb = max_hebb

            # Max cause episode group hebb to current recipient
            if analogs[driver].driver_episode is not None:
                analog.driver_episode = analogs[driver].driver_episode
                analog.driver_episode_act = analogs[driver].driver_episode_act
                hebbset = analogs[driver].driver_episode.get_hebbset(analog_name)
                max_hebb = 0.0
                if len(hebbset) > 0:
                    max_hebb = max(hebbs[i].perm_wt for i in hebbset)
                analog.driver_epiosde_hebb = max_hebb
                        
def traverse_tokens(analogs, driver, hebbs, hebb_bias, j, action_flag='map', analog_flag='driver'):
    '''Traverses the Token units in each analog and updates their inputs.  The input methods
    belong to the various Token unit classes.

    The calling code in cortrol.Firecontrol.run_mapping() has been changed so
    that only analogs that SHOULD be activated have an analog_type ==
    'recipient'.  All other analogs have an analog_type == 'driver' or None.
    Therefore, this code no longer checks for the analog_location/analog.name'''

    if analog_flag == 'driver':
        analog = analogs[driver]

        for pred in analog.preds:
            pred.get_td_input()
            pred.net_input = pred.td_input
            
        for obj in analog.objs:
            obj.get_td_input()
            obj.net_input = obj.td_input
            
        # No RF inputs in driver
        
        for prop in analog.props:
            prop.get_bu_input()
            prop.get_td_input()
            prop.get_driver_net_input()

        for group in analog.groups:
            group.get_driver_bu_input()
            group.get_driver_net_input()

    if analog_flag == 'recipient':
        for analog in analogs.itervalues():
            if analog.analog_type == 'recipient' and analog.top_down_ok == True:

                for pred in analog.preds:
                    # Get within_prop_input
                    pred.get_bu_input()
                    pred.get_td_input()
                    # Get within_class_input: NONE
                    # Get outprop_input from other RFs
                    pred.get_outprop_input(analog)
                    # Get total_hebb_input
                    pred.get_hebb_excitation(analog, driver, hebbs)
                    pred.get_max_hebb_w(driver, hebbs)
                    pred.get_hebb_inhib(analog)
                    pred.get_total_hebb_input(hebb_bias, action_flag)
                    # Get net_input
                    pred.get_recipient_net_input()
                    
                for obj in analog.objs:
                    # Get within_prop_input
                    obj.get_bu_input()
                    obj.get_td_input()
                    # Get within_class_input: NONE
                    # Get outprop_input from other RFs
                    obj.get_outprop_input(analog)
                    # Get total_hebb_input
                    obj.get_hebb_excitation(analog, driver, hebbs)
                    obj.get_max_hebb_w(driver, hebbs)
                    obj.get_hebb_inhib(analog)
                    obj.get_total_hebb_input(hebb_bias, action_flag)
                    # Get net_input
                    obj.get_recipient_net_input()

                for rf in analog.rfbindings:
                    # Get within_prop_input
                    rf.get_bu_input(action_flag)
                    rf.get_td_input()
                    # Get within_class_input
                    rf.get_within_class_input(analog)
                    # Get outprop_input from other props
                    rf.get_outprop_input(analog)
                    # Get total_hebb_input
                    rf.get_hebb_excitation(analog, driver, hebbs)
                    rf.get_max_hebb_w(driver, hebbs)
                    rf.get_hebb_inhib(analog)
                    rf.get_total_hebb_input(hebb_bias, action_flag)
                    # Get net_input
                    rf.get_recipient_net_input()
                    
                for prop in analog.props:
                    # Get within_prop_input
                    prop.get_bu_input()
                    prop.get_td_input()
                    # Get within_class_input
                    prop.get_within_class_input(analog)
                    # Get outprop_input from other props
                    prop.get_outprop_input(analog)
                    # Get total_hebb_input
                    prop.get_hebb_excitation(analog, driver, hebbs)
                    prop.get_max_hebb_w(driver, hebbs)
                    prop.get_hebb_inhib(analog)
                    prop.get_total_hebb_input(hebb_bias, action_flag)
                    # Get net_input
                    prop.get_recipient_net_input()

                for group in analog.groups:
                    # Get within_group_input
                    group.get_recip_bu_input(analog)
                    # Get within_class_input
                    group.get_within_class_input(analog)
                    group.get_outprop_input(analog)
                    # Get total_hebb_input
                    group.get_hebb_excitation(analog, driver, hebbs)
                    group.get_max_hebb_w(driver, hebbs)
                    group.get_hebb_inhib(analog)
                    group.get_total_hebb_input(hebb_bias, action_flag)
                    # Get net_input
                    group.get_recipient_net_input()

            if analog.analog_type == 'recipient' and analog.top_down_ok == False:
                
                for pred in analog.preds:
                    pred.get_bu_input()
                    pred.get_recipient_net_input()
                    
                for obj in analog.objs:
                    obj.get_bu_input()
                    obj.get_recipient_net_input()
                    
                for rf in analog.rfbindings:
                    rf.get_bu_input(action_flag)
                    rf.get_within_class_input(analog)
                    rf.get_recipient_net_input()
                    
                for prop in analog.props:
                    prop.get_bu_input()
                    prop.get_within_class_input(analog)
                    prop.get_recipient_net_input()

                for group in analog.groups:
                    group.get_recip_bu_input(analog)
                    group.get_within_class_input(analog)
                    group.get_recipient_net_input()

        
        
def set_recipient_mode(analogs, driver, hebbs, phi, j):
    '''Sets the mode of props in the recipient.'''
    
    for analog in analogs.itervalues():
        if analog.analog_type == 'recipient':
            for prop in analog.props:
                prop.get_mode_evidence(driver, hebbs)
                prop.setmode(phi)
                
def traverse_semantics(analogs, semantics):
    '''Returns net inputs and activations of semantic units.'''
    
    for semantic in semantics:
        semantic.get_semantic_net_input(analogs)
    for semantic in semantics:
        semantic.get_semantic_act(semantics)

def update_activations(analogs, gamma, delta, action_flag='map'):
    '''Update activations of all units in the system.

    As with traverse_tokens(), this function assumes that only the appropriate
    analogs have an analog_type == 'recipient', and all other analogs have
    analog_type == 'driver' or None.'''

    for analog in analogs.itervalues():
        if analog.analog_type == 'driver':
            for pred in analog.preds:
                pred.get_act_delta(gamma, delta)
                pred.update_act()
            for obj in analog.objs:
                obj.get_act_delta(gamma, delta)
                obj.update_act()
            for prop in analog.props:
                prop.get_act_delta(gamma, delta)
                prop.update_act()
            for group in analog.groups:
                group.get_act_delta(gamma, delta)
                group.update_act()
                
        elif analog.analog_type == 'recipient':
            for pred in analog.preds:
                pred.get_act_delta(gamma, delta)
                pred.update_act()
            for obj in analog.objs:
                obj.get_act_delta(gamma, delta)
                obj.update_act()
            for rf in analog.rfbindings:
                rf.get_act_delta(gamma, delta)
                rf.update_act()
            for prop in analog.props:
                prop.get_act_delta(gamma, delta)
                prop.update_act()
            for group in analog.groups:
                group.get_act_delta(gamma, delta)
                group.update_act()

def update_semantic_inhibitor(semantics):
    for semantic in semantics:
        semantic.net_input = 0.0
        semantic.act = 0.0
        semantic.delta = 0.0


